"""
Novell ASAM Runner
==================

.. versionadded:: 2015.8.0

Runner to interact with Novell ASAM Fan-Out Driver

:codeauthor: Nitin Madhok <nmadhok@g.clemson.edu>

To use this runner, set up the Novell Fan-Out Driver URL, username and password in the
master configuration at ``/etc/salt/master`` or ``/etc/salt/master.d/asam.conf``:

.. code-block:: yaml

    asam:
      prov1.domain.com
        username: "testuser"
        password: "verybadpass"
        verify_ssl: true
      prov2.domain.com
        username: "testuser"
        password: "verybadpass"
        verify_ssl: true

.. note::

    Optionally, ``protocol`` and ``port`` can be specified if the Fan-Out Driver server
    is not using the defaults. Default is ``protocol: https`` and ``port: 3451``.

"""

import logging

HAS_LIBS = False
try:
    import html.parser

    import requests

    HAS_LIBS = True

    # pylint: disable=abstract-method

    class ASAMHTMLParser(html.parser.HTMLParser):  # fix issue #30477
        def __init__(self):
            html.parser.HTMLParser.__init__(self)
            self.data = []

        def handle_starttag(self, tag, attrs):
            if tag != "a":
                return
            for attr in attrs:
                if attr[0] != "href":
                    return
                self.data.append(attr[1])

    # pylint: enable=abstract-method


except ImportError:
    pass

log = logging.getLogger(__name__)


def __virtual__():
    """
    Check for ASAM Fan-Out driver configuration in master config file
    or directory and load runner only if it is specified
    """
    if not HAS_LIBS:
        return False

    if _get_asam_configuration() is False:
        return False
    return True


def _get_asam_configuration(driver_url=""):
    """
    Return the configuration read from the master configuration
    file or directory
    """
    asam_config = __opts__["asam"] if "asam" in __opts__ else None

    if asam_config:
        try:
            for asam_server, service_config in asam_config.items():
                username = service_config.get("username", None)
                password = service_config.get("password", None)
                protocol = service_config.get("protocol", "https")
                port = service_config.get("port", 3451)
                verify_ssl = service_config.get("verify_ssl")

                if verify_ssl is None:
                    verify_ssl = True

                if not username or not password:
                    log.error(
                        "Username or Password has not been specified in the "
                        "master configuration for %s",
                        asam_server,
                    )
                    return False

                ret = {
                    "platform_edit_url": "{}://{}:{}/config/PlatformEdit.html".format(
                        protocol, asam_server, port
                    ),
                    "platform_config_url": (
                        "{}://{}:{}/config/PlatformConfig.html".format(
                            protocol, asam_server, port
                        )
                    ),
                    "platformset_edit_url": (
                        "{}://{}:{}/config/PlatformSetEdit.html".format(
                            protocol, asam_server, port
                        )
                    ),
                    "platformset_config_url": (
                        "{}://{}:{}/config/PlatformSetConfig.html".format(
                            protocol, asam_server, port
                        )
                    ),
                    "username": username,
                    "password": password,
                    "verify_ssl": verify_ssl,
                }

                if (not driver_url) or (driver_url == asam_server):
                    return ret
        except Exception as exc:  # pylint: disable=broad-except
            log.error("Exception encountered: %s", exc)
            return False

        if driver_url:
            log.error(
                "Configuration for %s has not been specified in the master "
                "configuration",
                driver_url,
            )
            return False

    return False


def _make_post_request(url, data, auth, verify=True):
    r = requests.post(url, data=data, auth=auth, verify=verify)
    if r.status_code != requests.codes.ok:
        r.raise_for_status()
    else:
        return r.text.split("\n")


def _parse_html_content(html_content):
    parser = ASAMHTMLParser()
    for line in html_content:
        if line.startswith("<META"):
            html_content.remove(line)
        else:
            parser.feed(line)

    return parser


def _get_platformset_name(data, platform_name):
    for item in data:
        if platform_name in item and item.startswith("PlatformEdit.html?"):
            parameter_list = item.split("&")
            for parameter in parameter_list:
                if parameter.startswith("platformSetName"):
                    return parameter.split("=")[1]

    return None


def _get_platforms(data):
    platform_list = []
    for item in data:
        if item.startswith("PlatformEdit.html?"):
            parameter_list = item.split("PlatformEdit.html?", 1)[1].split("&")
            for parameter in parameter_list:
                if parameter.startswith("platformName"):
                    platform_list.append(parameter.split("=")[1])

    return platform_list


def _get_platform_sets(data):
    platform_set_list = []
    for item in data:
        if item.startswith("PlatformSetEdit.html?"):
            parameter_list = item.split("PlatformSetEdit.html?", 1)[1].split("&")
            for parameter in parameter_list:
                if parameter.startswith("platformSetName"):
                    platform_set_list.append(
                        parameter.split("=")[1].replace("%20", " ")
                    )

    return platform_set_list


def remove_platform(name, server_url):
    """
    To remove specified ASAM platform from the Novell Fan-Out Driver

    CLI Example:

    .. code-block:: bash

        salt-run asam.remove_platform my-test-vm prov1.domain.com
    """
    config = _get_asam_configuration(server_url)
    if not config:
        return False

    url = config["platform_config_url"]

    data = {
        "manual": "false",
    }

    auth = (config["username"], config["password"])

    try:
        html_content = _make_post_request(url, data, auth, verify=config["verify_ssl"])
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = "Failed to look up existing platforms on {}".format(server_url)
        log.error("%s:\n%s", err_msg, exc)
        return {name: err_msg}

    parser = _parse_html_content(html_content)
    platformset_name = _get_platformset_name(parser.data, name)

    if platformset_name:
        log.debug(platformset_name)
        data["platformName"] = name
        data["platformSetName"] = str(platformset_name)
        data["postType"] = "platformRemove"
        data["Submit"] = "Yes"
        try:
            html_content = _make_post_request(
                url, data, auth, verify=config["verify_ssl"]
            )
        except Exception as exc:  # pylint: disable=broad-except
            err_msg = "Failed to delete platform from {}".format(server_url)
            log.error("%s:\n%s", err_msg, exc)
            return {name: err_msg}

        parser = _parse_html_content(html_content)
        platformset_name = _get_platformset_name(parser.data, name)
        if platformset_name:
            return {name: "Failed to delete platform from {}".format(server_url)}
        else:
            return {name: "Successfully deleted platform from {}".format(server_url)}
    else:
        return {name: "Specified platform name does not exist on {}".format(server_url)}


def list_platforms(server_url):
    """
    To list all ASAM platforms present on the Novell Fan-Out Driver

    CLI Example:

    .. code-block:: bash

        salt-run asam.list_platforms prov1.domain.com
    """
    config = _get_asam_configuration(server_url)
    if not config:
        return False

    url = config["platform_config_url"]

    data = {
        "manual": "false",
    }

    auth = (config["username"], config["password"])

    try:
        html_content = _make_post_request(url, data, auth, verify=config["verify_ssl"])
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = "Failed to look up existing platforms"
        log.error("%s:\n%s", err_msg, exc)
        return {server_url: err_msg}

    parser = _parse_html_content(html_content)
    platform_list = _get_platforms(parser.data)

    if platform_list:
        return {server_url: platform_list}
    else:
        return {server_url: "No existing platforms found"}


def list_platform_sets(server_url):
    """
    To list all ASAM platform sets present on the Novell Fan-Out Driver

    CLI Example:

    .. code-block:: bash

        salt-run asam.list_platform_sets prov1.domain.com
    """
    config = _get_asam_configuration(server_url)
    if not config:
        return False

    url = config["platformset_config_url"]

    data = {
        "manual": "false",
    }

    auth = (config["username"], config["password"])

    try:
        html_content = _make_post_request(url, data, auth, verify=config["verify_ssl"])
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = "Failed to look up existing platform sets"
        log.error("%s:\n%s", err_msg, exc)
        return {server_url: err_msg}

    parser = _parse_html_content(html_content)
    platform_set_list = _get_platform_sets(parser.data)

    if platform_set_list:
        return {server_url: platform_set_list}
    else:
        return {server_url: "No existing platform sets found"}


def add_platform(name, platform_set, server_url):
    """
    To add an ASAM platform using the specified ASAM platform set on the Novell
    Fan-Out Driver

    CLI Example:

    .. code-block:: bash

        salt-run asam.add_platform my-test-vm test-platform-set prov1.domain.com
    """
    config = _get_asam_configuration(server_url)
    if not config:
        return False

    platforms = list_platforms(server_url)
    if name in platforms[server_url]:
        return {name: "Specified platform already exists on {}".format(server_url)}

    platform_sets = list_platform_sets(server_url)
    if platform_set not in platform_sets[server_url]:
        return {name: "Specified platform set does not exist on {}".format(server_url)}

    url = config["platform_edit_url"]

    data = {
        "platformName": name,
        "platformSetName": platform_set,
        "manual": "false",
        "previousURL": "/config/platformAdd.html",
        "postType": "PlatformAdd",
        "Submit": "Apply",
    }

    auth = (config["username"], config["password"])

    try:
        html_content = _make_post_request(url, data, auth, verify=config["verify_ssl"])
    except Exception as exc:  # pylint: disable=broad-except
        err_msg = "Failed to add platform on {}".format(server_url)
        log.error("%s:\n%s", err_msg, exc)
        return {name: err_msg}

    platforms = list_platforms(server_url)
    if name in platforms[server_url]:
        return {name: "Successfully added platform on {}".format(server_url)}
    else:
        return {name: "Failed to add platform on {}".format(server_url)}



import pytest

import salt.client.ssh.client
import salt.utils.msgpack
from salt.client import ssh
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True),
    pytest.mark.slow_test,
]


@pytest.fixture
def opts(tmp_path, temp_salt_master):
    updated_values = {
        "argv": [
            "ssh.set_auth_key",
            "root",
            "hobn+amNAXSBTiOXEqlBjGB...rsa root@master",
        ],
        "__role": "master",
        "cachedir": str(tmp_path),
        "extension_modules": str(tmp_path / "extmods"),
        "selected_target_option": "glob",
    }

    opts = temp_salt_master.config.copy()
    opts.update(updated_values)
    return opts


@pytest.fixture
def target():
    return {
        "passwd": "abc123",
        "ssh_options": None,
        "sudo": False,
        "identities_only": False,
        "host": "login1",
        "user": "root",
        "timeout": 65,
        "remote_port_forwards": None,
        "sudo_user": "",
        "port": "22",
        "priv": "/etc/salt/pki/master/ssh/salt-ssh.rsa",
    }


@pytest.fixture
def roster():
    return """
        localhost:
          host: 127.0.0.1
          port: 2827
        """


@pytest.mark.parametrize(
    "test_opts",
    [
        ("extra_filerefs", "salt://foobar", True),
        ("host", "testhost", False),
        ("ssh_user", "testuser", True),
        ("ssh_passwd", "testpasswd", True),
        ("ssh_port", 23, False),
        ("ssh_sudo", True, True),
        ("ssh_sudo_user", "sudouser", False),
        ("ssh_priv", "test_priv", True),
        ("ssh_priv_passwd", "sshpasswd", True),
        ("ssh_identities_only", True, True),
        ("ssh_remote_port_forwards", "test", True),
        ("ssh_options", ["test1", "test2"], True),
        ("ssh_max_procs", 2, True),
        ("ssh_askpass", True, True),
        ("ssh_key_deploy", True, True),
        ("ssh_update_roster", True, True),
        ("ssh_scan_ports", "test", True),
        ("ssh_scan_timeout", 1.0, True),
        ("ssh_timeout", 1, False),
        ("ssh_log_file", "/tmp/test", True),
        ("raw_shell", True, True),
        ("refresh_cache", True, True),
        ("roster", "/test", True),
        ("roster_file", "/test1", True),
        ("rosters", ["test1"], False),
        ("ignore_host_keys", True, True),
        ("min_extra_mods", "test", True),
        ("thin_extra_mods", "test1", True),
        ("verbose", True, True),
        ("static", True, True),
        ("ssh_wipe", True, True),
        ("rand_thin_dir", True, True),
        ("regen_thin", True, True),
        ("ssh_run_pre_flight", True, True),
        ("no_host_keys", True, True),
        ("saltfile", "/tmp/test", True),
        ("doesnotexist", None, False),
    ],
)
def test_ssh_kwargs(test_opts):
    """
    test all ssh kwargs are not excluded from kwargs
    when preparing the SSH opts
    """
    opt_key = test_opts[0]
    opt_value = test_opts[1]
    # Is the kwarg in salt.utils.parsers?
    in_parser = test_opts[2]

    opts = {
        "eauth": "auto",
        "username": "test",
        "password": "test",
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        opt_key: opt_value,
    }
    client = salt.client.ssh.client.SSHClient(disable_custom_roster=True)
    if in_parser:
        ssh_kwargs = salt.utils.parsers.SaltSSHOptionParser().defaults
        assert opt_key in ssh_kwargs

    with patch("salt.roster.get_roster_file", MagicMock(return_value="")), patch(
        "salt.client.ssh.shell.gen_key"
    ), patch("salt.fileserver.Fileserver.update"), patch("salt.utils.thin.gen_thin"):
        ssh_obj = client._prep_ssh(**opts)
        assert ssh_obj.opts.get(opt_key, None) == opt_value


@pytest.mark.slow_test
def test_expand_target_ip_address(opts, roster):
    """
    test expand_target when target is root@<ip address>
    """
    host = "127.0.0.1"
    user = "test-user@"
    opts["tgt"] = user + host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == user + host
    with patch(
        "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
    ), patch(
        "salt.client.ssh.compile_template",
        MagicMock(return_value=salt.utils.yaml.safe_load(roster)),
    ):
        client._expand_target()
    assert opts["tgt"] == host


def test_expand_target_no_host(opts, tmp_path):
    """
    test expand_target when host is not included in the rosterdata
    """
    host = "127.0.0.1"
    user = "test-user@"
    opts["tgt"] = user + host

    roster = """
        localhost: 127.0.0.1
        """
    roster_file = str(tmp_path / "test_roster_no_host")
    with salt.utils.files.fopen(roster_file, "w") as fp:
        salt.utils.yaml.safe_dump(salt.utils.yaml.safe_load(roster), fp)

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == user + host
    with patch("salt.roster.get_roster_file", MagicMock(return_value=roster_file)):
        client._expand_target()
    assert opts["tgt"] == host


def test_expand_target_dns(opts, roster):
    """
    test expand_target when target is root@<dns>
    """
    host = "localhost"
    user = "test-user@"
    opts["tgt"] = user + host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == user + host
    with patch(
        "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
    ), patch(
        "salt.client.ssh.compile_template",
        MagicMock(return_value=salt.utils.yaml.safe_load(roster)),
    ):
        client._expand_target()
    assert opts["tgt"] == host


def test_expand_target_no_user(opts, roster):
    """
    test expand_target when no user defined
    """
    host = "127.0.0.1"
    opts["tgt"] = host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == host

    with patch(
        "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
    ), patch(
        "salt.client.ssh.compile_template",
        MagicMock(return_value=salt.utils.yaml.safe_load(roster)),
    ):
        client._expand_target()
    assert opts["tgt"] == host


def test_update_targets_ip_address(opts):
    """
    test update_targets when host is ip address
    """
    host = "127.0.0.1"
    user = "test-user@"
    opts["tgt"] = user + host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == user + host
    client._update_targets()
    assert opts["tgt"] == host
    assert client.targets[host]["user"] == user.split("@")[0]


def test_update_targets_dns(opts):
    """
    test update_targets when host is dns
    """
    host = "localhost"
    user = "test-user@"
    opts["tgt"] = user + host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == user + host
    client._update_targets()
    assert opts["tgt"] == host
    assert client.targets[host]["user"] == user.split("@")[0]


def test_update_targets_no_user(opts):
    """
    test update_targets when no user defined
    """
    host = "127.0.0.1"
    opts["tgt"] = host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == host
    client._update_targets()
    assert opts["tgt"] == host


def test_update_expand_target_dns(opts, roster):
    """
    test update_targets and expand_target when host is dns
    """
    host = "localhost"
    user = "test-user@"
    opts["tgt"] = user + host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        client = ssh.SSH(opts)
    assert opts["tgt"] == user + host
    with patch(
        "salt.roster.get_roster_file", MagicMock(return_value="/etc/salt/roster")
    ), patch(
        "salt.client.ssh.compile_template",
        MagicMock(return_value=salt.utils.yaml.safe_load(roster)),
    ):
        client._expand_target()
    client._update_targets()
    assert opts["tgt"] == host
    assert client.targets[host]["user"] == user.split("@")[0]


def test_parse_tgt(opts):
    """
    test parse_tgt when user and host set on
    the ssh cli tgt
    """
    host = "localhost"
    user = "test-user@"
    opts["tgt"] = user + host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        assert not opts.get("ssh_cli_tgt")
        client = ssh.SSH(opts)
        assert client.parse_tgt["hostname"] == host
        assert client.parse_tgt["user"] == user.split("@")[0]
        assert opts.get("ssh_cli_tgt") == user + host


def test_parse_tgt_no_user(opts):
    """
    test parse_tgt when only the host set on
    the ssh cli tgt
    """
    host = "localhost"
    opts["ssh_user"] = "ssh-usr"
    opts["tgt"] = host

    with patch("salt.utils.network.is_reachable_host", MagicMock(return_value=False)):
        assert not opts.get("ssh_cli_tgt")
        client = ssh.SSH(opts)
        assert client.parse_tgt["hostname"] == host
        assert client.parse_tgt["user"] == opts["ssh_user"]
        assert opts.get("ssh_cli_tgt") == host


def test_extra_filerefs(tmp_path, opts):
    """
    test "extra_filerefs" are not excluded from kwargs
    when preparing the SSH opts
    """
    ssh_opts = {
        "eauth": "auto",
        "username": "test",
        "password": "test",
        "client": "ssh",
        "tgt": "localhost",
        "fun": "test.ping",
        "ssh_port": 22,
        "extra_filerefs": "salt://foobar",
    }
    roster = str(tmp_path / "roster")
    client = salt.client.ssh.client.SSHClient(mopts=opts, disable_custom_roster=True)
    with patch("salt.roster.get_roster_file", MagicMock(return_value=roster)):
        ssh_obj = client._prep_ssh(**ssh_opts)
        assert ssh_obj.opts.get("extra_filerefs", None) == "salt://foobar"


def test_key_deploy_permission_denied_scp(tmp_path, opts):
    """
    test "key_deploy" function when
    permission denied authentication error
    when attempting to use scp to copy file
    to target
    """
    host = "localhost"
    passwd = "password"
    usr = "ssh-usr"
    opts["ssh_user"] = usr
    opts["tgt"] = host

    ssh_ret = {
        host: {
            "stdout": "\rroot@192.168.1.187's password: \n\rroot@192.168.1.187's password: \n\rroot@192.168.1.187's password: \n",
            "stderr": "Permission denied, please try again.\nPermission denied, please try again.\nroot@192.168.1.187: Permission denied (publickey,gssapi-keyex,gssapi-with-micimport pudb; pu.dbassword).\nscp: Connection closed\n",
            "retcode": 255,
        }
    }
    key_run_ret = {
        "localhost": {
            "jid": "20230922155652279959",
            "return": "test",
            "retcode": 0,
            "id": "test",
            "fun": "cmd.run",
            "fun_args": ["echo test"],
        }
    }
    patch_roster_file = patch("salt.roster.get_roster_file", MagicMock(return_value=""))
    with patch_roster_file:
        client = ssh.SSH(opts)
    patch_input = patch("builtins.input", side_effect=["y"])
    patch_getpass = patch("getpass.getpass", return_value=["password"])
    mock_key_run = MagicMock(return_value=key_run_ret)
    patch_key_run = patch("salt.client.ssh.SSH._key_deploy_run", mock_key_run)
    with patch_input, patch_getpass, patch_key_run:
        ret = client.key_deploy(host, ssh_ret)
    assert mock_key_run.call_args_list[0][0] == (
        host,
        {"passwd": [passwd], "host": host, "user": usr},
        True,
    )
    assert ret == key_run_ret
    assert mock_key_run.call_count == 1


def test_key_deploy_permission_denied_file_scp(tmp_path, opts):
    """
    test "key_deploy" function when permission denied
    due to not having access to copy the file to the target
    We do not want to deploy the key, because this is not
    an authentication to the target error.
    """
    host = "localhost"
    passwd = "password"
    usr = "ssh-usr"
    opts["ssh_user"] = usr
    opts["tgt"] = host

    mock_key_run = MagicMock(return_value=False)
    patch_key_run = patch("salt.client.ssh.SSH._key_deploy_run", mock_key_run)

    ssh_ret = {
        "localhost": {
            "stdout": "",
            "stderr": 'scp: dest open "/tmp/preflight.sh": Permission denied\nscp: failed to upload file /etc/salt/preflight.sh to /tmp/preflight.sh\n',
            "retcode": 1,
        }
    }
    patch_roster_file = patch("salt.roster.get_roster_file", MagicMock(return_value=""))
    with patch_roster_file:
        client = ssh.SSH(opts)
    ret = client.key_deploy(host, ssh_ret)
    assert ret == ssh_ret
    assert mock_key_run.call_count == 0


def test_key_deploy_no_permission_denied(tmp_path, opts):
    """
    test "key_deploy" function when no permission denied
    is returned
    """
    host = "localhost"
    passwd = "password"
    usr = "ssh-usr"
    opts["ssh_user"] = usr
    opts["tgt"] = host

    mock_key_run = MagicMock(return_value=False)
    patch_key_run = patch("salt.client.ssh.SSH._key_deploy_run", mock_key_run)
    ssh_ret = {
        "localhost": {
            "jid": "20230922161937998385",
            "return": "test",
            "retcode": 0,
            "id": "test",
            "fun": "cmd.run",
            "fun_args": ["echo test"],
        }
    }
    patch_roster_file = patch("salt.roster.get_roster_file", MagicMock(return_value=""))
    with patch_roster_file:
        client = ssh.SSH(opts)
    ret = client.key_deploy(host, ssh_ret)
    assert ret == ssh_ret
    assert mock_key_run.call_count == 0



import types

import pytest

from salt.cloud.clouds import azurearm as azure
from tests.support.mock import MagicMock, create_autospec, patch


def copy_func(func, globals=None):
    # I do not know that this is complete, but it's sufficient for now.
    # The key to "moving" the function to another module (or stubbed module)
    # is to update __globals__.

    copied_func = types.FunctionType(
        func.__code__, globals, func.__name__, func.__defaults__, func.__closure__
    )
    copied_func.__module__ = func.__module__
    copied_func.__doc__ = func.__doc__
    copied_func.__kwdefaults__ = func.__kwdefaults__
    copied_func.__dict__.update(func.__dict__)
    return copied_func


def mock_module(mod, sut=None):
    if sut is None:
        sut = [None]

    mock = create_autospec(mod)

    # we need to provide a '__globals__' so functions being tested behave correctly.
    mock_globals = {}

    # exclude the system under test
    for name in sut:
        attr = getattr(mod, name)
        if isinstance(attr, types.FunctionType):
            attr = copy_func(attr, mock_globals)
        setattr(mock, name, attr)

    # fully populate our mock_globals
    for name in mod.__dict__:
        if name in mock.__dict__:
            mock_globals[name] = mock.__dict__[name]
        elif type(getattr(mod, name)) is type(types):  # is a module
            mock_globals[name] = getattr(mock, name)
        else:
            mock_globals[name] = mod.__dict__[name]

    return mock


@pytest.fixture
def configure_loader_modules():
    return {azure: {"__opts__": {}, "__active_provider_name__": None}}


@pytest.mark.skipif(not azure.HAS_LIBS, reason="azure not available")
def test_function_signatures():
    mock_azure = mock_module(azure, sut=["request_instance", "__opts__", "__utils__"])
    mock_azure.create_network_interface.return_value = [
        MagicMock(),
        MagicMock(),
        MagicMock(),
    ]
    mock_azure.salt.utils.stringutils.to_str.return_value = "P4ssw0rd"
    mock_azure.salt.utils.cloud.gen_keys.return_value = [MagicMock(), MagicMock()]
    mock_azure.__opts__["pki_dir"] = None

    mock_azure.request_instance.__globals__[
        "__builtins__"
    ] = mock_azure.request_instance.__globals__["__builtins__"].copy()
    mock_azure.request_instance.__globals__["__builtins__"]["getattr"] = MagicMock()

    mock_azure.__utils__["cloud.fire_event"] = mock_azure.salt.utils.cloud.fire_event
    mock_azure.__utils__[
        "cloud.filter_event"
    ] = mock_azure.salt.utils.cloud.filter_event
    mock_azure.__opts__["sock_dir"] = MagicMock()
    mock_azure.__opts__["transport"] = MagicMock()

    mock_azure.request_instance(
        {"image": "http://img", "storage_account": "blah", "size": ""}
    )

    # we literally only check that a final creation call occurred.
    mock_azure.get_conn.return_value.virtual_machines.create_or_update.assert_called_once()


def test_get_configured_provider():
    mock_azure = mock_module(
        azure, sut=["get_configured_provider", "__opts__", "__utils__"]
    )

    good_combos = [
        {
            "subscription_id": "3287abc8-f98a-c678-3bde-326766fd3617",
            "tenant": "ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF",
            "client_id": "ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF",
            "secret": "XXXXXXXXXXXXXXXXXXXXXXXX",
        },
        {
            "subscription_id": "3287abc8-f98a-c678-3bde-326766fd3617",
            "username": "larry",
            "password": "123pass",
        },
        {"subscription_id": "3287abc8-f98a-c678-3bde-326766fd3617"},
    ]

    for combo in good_combos:
        mock_azure.__opts__["providers"] = {"azure_test": {"azurearm": combo}}
        assert azure.get_configured_provider() == combo

    bad_combos = [
        {"subscrption": "3287abc8-f98a-c678-3bde-326766fd3617"},
        {},
    ]

    for combo in bad_combos:
        mock_azure.__opts__["providers"] = {"azure_test": {"azurearm": combo}}
        assert not azure.get_configured_provider()


def test_get_conn():
    mock_azure = mock_module(azure, sut=["get_conn", "__opts__", "__utils__"])

    mock_azure.__opts__["providers"] = {
        "azure_test": {
            "azurearm": {
                "subscription_id": "3287abc8-f98a-c678-3bde-326766fd3617",
                "driver": "azurearm",
                "password": "monkeydonkey",
            }
        }
    }
    # password is stripped if username not provided
    expected = {"subscription_id": "3287abc8-f98a-c678-3bde-326766fd3617"}
    with patch(
        "salt.utils.azurearm.get_client", side_effect=lambda client_type, **kw: kw
    ):
        assert azure.get_conn(client_type="compute") == expected

    mock_azure.__opts__["providers"] = {
        "azure_test": {
            "azurearm": {
                "subscription_id": "3287abc8-f98a-c678-3bde-326766fd3617",
                "driver": "azurearm",
                "username": "donkeymonkey",
                "password": "monkeydonkey",
            }
        }
    }
    # username and password via provider config
    expected = {
        "subscription_id": "3287abc8-f98a-c678-3bde-326766fd3617",
        "username": "donkeymonkey",
        "password": "monkeydonkey",
    }
    with patch(
        "salt.utils.azurearm.get_client", side_effect=lambda client_type, **kw: kw
    ):
        assert azure.get_conn(client_type="compute") == expected



"""
Return data to an elasticsearch server for indexing.

:maintainer:    Jurnell Cockhren <jurnell.cockhren@sophicware.com>, Arnold Bechtoldt <mail@arnoldbechtoldt.com>
:maturity:      New
:depends:       `elasticsearch-py <https://elasticsearch-py.readthedocs.io/en/latest/>`_
:platform:      all

To enable this returner the elasticsearch python client must be installed
on the desired minions (all or some subset).

Please see documentation of :mod:`elasticsearch execution module <salt.modules.elasticsearch>`
for a valid connection configuration.

.. warning::

        The index that you wish to store documents will be created by Elasticsearch automatically if
        doesn't exist yet. It is highly recommended to create predefined index templates with appropriate mapping(s)
        that will be used by Elasticsearch upon index creation. Otherwise you will have problems as described in #20826.

To use the returner per salt call:

.. code-block:: bash

    salt '*' test.ping --return elasticsearch

In order to have the returner apply to all minions:

.. code-block:: yaml

    ext_job_cache: elasticsearch

Minion configuration:
    debug_returner_payload': False
        Output the payload being posted to the log file in debug mode

    doc_type: 'default'
        Document type to use for normal return messages

    functions_blacklist
        Optional list of functions that should not be returned to elasticsearch

    index_date: False
        Use a dated index (e.g. <index>-2016.11.29)

    master_event_index: 'salt-master-event-cache'
        Index to use when returning master events

    master_event_doc_type: 'efault'
        Document type to use got master events

    master_job_cache_index: 'salt-master-job-cache'
        Index to use for master job cache

    master_job_cache_doc_type: 'default'
        Document type to use for master job cache

    number_of_shards: 1
        Number of shards to use for the indexes

    number_of_replicas: 0
        Number of replicas to use for the indexes

    NOTE: The following options are valid for 'state.apply', 'state.sls' and 'state.highstate' functions only.

    states_count: False
        Count the number of states which succeeded or failed and return it in top-level item called 'counts'.
        States reporting None (i.e. changes would be made but it ran in test mode) are counted as successes.
    states_order_output: False
        Prefix the state UID (e.g. file_|-yum_configured_|-/etc/yum.conf_|-managed) with a zero-padded version
        of the '__run_num__' value to allow for easier sorting. Also store the state function (i.e. file.managed)
        into a new key '_func'. Change the index to be '<index>-ordered' (e.g. salt-state_apply-ordered).
    states_single_index: False
        Store results for state.apply, state.sls and state.highstate in the salt-state_apply index
        (or -ordered/-<date>) indexes if enabled

.. code-block:: yaml

    elasticsearch:
        hosts:
          - "10.10.10.10:9200"
          - "10.10.10.11:9200"
          - "10.10.10.12:9200"
        index_date: True
        number_of_shards: 5
        number_of_replicas: 1
        debug_returner_payload: True
        states_count: True
        states_order_output: True
        states_single_index: True
        functions_blacklist:
          - test.ping
          - saltutil.find_job
"""


import datetime
import logging
import uuid
from datetime import timedelta, tzinfo

import salt.returners
import salt.utils.jid
import salt.utils.json

__virtualname__ = "elasticsearch"

log = logging.getLogger(__name__)

STATE_FUNCTIONS = {
    "state.apply": "state_apply",
    "state.highstate": "state_apply",
    "state.sls": "state_apply",
}


def __virtual__():
    if "elasticsearch.index_exists" not in __salt__:
        return (
            False,
            "Elasticsearch module not availble.  Check that the elasticsearch library"
            " is installed.",
        )
    return __virtualname__


def _get_options(ret=None):
    """
    Get the returner options from salt.
    """

    defaults = {
        "debug_returner_payload": False,
        "doc_type": "default",
        "functions_blacklist": [],
        "index_date": False,
        "master_event_index": "salt-master-event-cache",
        "master_event_doc_type": "default",
        "master_job_cache_index": "salt-master-job-cache",
        "master_job_cache_doc_type": "default",
        "number_of_shards": 1,
        "number_of_replicas": 0,
        "states_order_output": False,
        "states_count": False,
        "states_single_index": False,
    }

    attrs = {
        "debug_returner_payload": "debug_returner_payload",
        "doc_type": "doc_type",
        "functions_blacklist": "functions_blacklist",
        "index_date": "index_date",
        "master_event_index": "master_event_index",
        "master_event_doc_type": "master_event_doc_type",
        "master_job_cache_index": "master_job_cache_index",
        "master_job_cache_doc_type": "master_job_cache_doc_type",
        "number_of_shards": "number_of_shards",
        "number_of_replicas": "number_of_replicas",
        "states_count": "states_count",
        "states_order_output": "states_order_output",
        "states_single_index": "states_single_index",
    }

    _options = salt.returners.get_returner_options(
        __virtualname__,
        ret,
        attrs,
        __salt__=__salt__,
        __opts__=__opts__,
        defaults=defaults,
    )
    return _options


def _ensure_index(index):
    index_exists = __salt__["elasticsearch.index_exists"](index)
    if not index_exists:
        options = _get_options()

        index_definition = {
            "settings": {
                "number_of_shards": options["number_of_shards"],
                "number_of_replicas": options["number_of_replicas"],
            }
        }
        __salt__["elasticsearch.index_create"]("{}-v1".format(index), index_definition)
        __salt__["elasticsearch.alias_create"]("{}-v1".format(index), index)


def _convert_keys(data):
    if isinstance(data, dict):
        new_data = {}
        for k, sub_data in data.items():
            if "." in k:
                new_data["_orig_key"] = k
                k = k.replace(".", "_")
            new_data[k] = _convert_keys(sub_data)
    elif isinstance(data, list):
        new_data = []
        for item in data:
            new_data.append(_convert_keys(item))
    else:
        return data

    return new_data


def returner(ret):
    """
    Process the return from Salt
    """

    job_fun = ret["fun"]
    job_fun_escaped = job_fun.replace(".", "_")
    job_id = ret["jid"]
    job_retcode = ret.get("retcode", 1)
    job_success = True if not job_retcode else False

    options = _get_options(ret)

    if job_fun in options["functions_blacklist"]:
        log.info(
            "Won't push new data to Elasticsearch, job with jid=%s and "
            "function=%s which is in the user-defined list of ignored "
            "functions",
            job_id,
            job_fun,
        )
        return
    if ret.get("data", None) is None and ret.get("return") is None:
        log.info(
            "Won't push new data to Elasticsearch, job with jid=%s was not successful",
            job_id,
        )
        return

    # Build the index name
    if options["states_single_index"] and job_fun in STATE_FUNCTIONS:
        index = "salt-{}".format(STATE_FUNCTIONS[job_fun])
    else:
        index = "salt-{}".format(job_fun_escaped)

    if options["index_date"]:
        index = "{}-{}".format(index, datetime.date.today().strftime("%Y.%m.%d"))

    counts = {}

    # Do some special processing for state returns
    if job_fun in STATE_FUNCTIONS:
        # Init the state counts
        if options["states_count"]:
            counts = {
                "succeeded": 0,
                "failed": 0,
            }

        # Prepend each state execution key in ret['return'] with a zero-padded
        # version of the '__run_num__' field allowing the states to be ordered
        # more easily. Change the index to be
        # index to be '<index>-ordered' so as not to clash with the unsorted
        # index data format
        if options["states_order_output"] and isinstance(ret["return"], dict):
            index = "{}-ordered".format(index)
            max_chars = len(str(len(ret["return"])))

            for uid, data in ret["return"].items():
                # Skip keys we've already prefixed
                if uid.startswith(tuple("0123456789")):
                    continue

                # Store the function being called as it's a useful key to search
                decoded_uid = uid.split("_|-")
                ret["return"][uid]["_func"] = "{}.{}".format(
                    decoded_uid[0], decoded_uid[-1]
                )

                # Prefix the key with the run order so it can be sorted
                new_uid = "{}_|-{}".format(
                    str(data["__run_num__"]).zfill(max_chars),
                    uid,
                )

                ret["return"][new_uid] = ret["return"].pop(uid)

        # Catch a state output that has failed and where the error message is
        # not in a dict as expected. This prevents elasticsearch from
        # complaining about a mapping error
        elif not isinstance(ret["return"], dict):
            ret["return"] = {"return": ret["return"]}

        # Need to count state successes and failures
        if options["states_count"]:
            for state_data in ret["return"].values():
                if state_data["result"] is False:
                    counts["failed"] += 1
                else:
                    counts["succeeded"] += 1

    # Ensure the index exists
    _ensure_index(index)

    # Build the payload
    class UTC(tzinfo):
        def utcoffset(self, dt):
            return timedelta(0)

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return timedelta(0)

    utc = UTC()
    data = {
        "@timestamp": datetime.datetime.now(utc).isoformat(),
        "success": job_success,
        "retcode": job_retcode,
        "minion": ret["id"],
        "fun": job_fun,
        "jid": job_id,
        "counts": counts,
        "data": _convert_keys(ret["return"]),
    }

    if options["debug_returner_payload"]:
        log.debug("elasicsearch payload: %s", data)

    # Post the payload
    ret = __salt__["elasticsearch.document_create"](
        index=index, doc_type=options["doc_type"], body=salt.utils.json.dumps(data)
    )


def event_return(events):
    """
    Return events to Elasticsearch

    Requires that the `event_return` configuration be set in master config.
    """
    options = _get_options()

    index = options["master_event_index"]
    doc_type = options["master_event_doc_type"]

    if options["index_date"]:
        index = "{}-{}".format(index, datetime.date.today().strftime("%Y.%m.%d"))

    _ensure_index(index)

    for event in events:
        data = {"tag": event.get("tag", ""), "data": event.get("data", "")}

    ret = __salt__["elasticsearch.document_create"](
        index=index,
        doc_type=doc_type,
        id=uuid.uuid4(),
        body=salt.utils.json.dumps(data),
    )


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)


def save_load(jid, load, minions=None):
    """
    Save the load to the specified jid id

    .. versionadded:: 2015.8.1
    """
    options = _get_options()

    index = options["master_job_cache_index"]
    doc_type = options["master_job_cache_doc_type"]

    _ensure_index(index)

    data = {
        "jid": jid,
        "load": load,
    }

    ret = __salt__["elasticsearch.document_create"](
        index=index, doc_type=doc_type, id=jid, body=salt.utils.json.dumps(data)
    )


def get_load(jid):
    """
    Return the load data that marks a specified jid

    .. versionadded:: 2015.8.1
    """
    options = _get_options()

    index = options["master_job_cache_index"]
    doc_type = options["master_job_cache_doc_type"]

    data = __salt__["elasticsearch.document_get"](
        index=index, id=jid, doc_type=doc_type
    )
    if data:
        return salt.utils.json.loads(data)
    return {}



"""
Provide authentication using simple LDAP binds

:depends:   - ldap Python module
"""
import itertools
import logging

from jinja2 import Environment

import salt.utils.data
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

try:
    # pylint: disable=no-name-in-module
    import ldap
    import ldap.filter
    import ldap.modlist

    HAS_LDAP = True
    # pylint: enable=no-name-in-module
except ImportError:
    HAS_LDAP = False

# Defaults, override in master config
__defopts__ = {
    "auth.ldap.basedn": "",
    "auth.ldap.uri": "",
    "auth.ldap.server": "localhost",
    "auth.ldap.port": "389",
    "auth.ldap.starttls": False,
    "auth.ldap.tls": False,
    "auth.ldap.no_verify": False,
    "auth.ldap.anonymous": False,
    "auth.ldap.scope": 2,
    "auth.ldap.groupou": "Groups",
    "auth.ldap.accountattributename": "memberUid",
    "auth.ldap.groupattribute": "memberOf",
    "auth.ldap.persontype": "person",
    "auth.ldap.groupclass": "posixGroup",
    "auth.ldap.activedirectory": False,
    "auth.ldap.freeipa": False,
    "auth.ldap.minion_stripdomains": [],
}


def _config(key, mandatory=True, opts=None):
    """
    Return a value for 'name' from master config file options or defaults.
    """
    try:
        if opts:
            value = opts["auth.ldap.{}".format(key)]
        else:
            value = __opts__["auth.ldap.{}".format(key)]
    except KeyError:
        try:
            value = __defopts__["auth.ldap.{}".format(key)]
        except KeyError:
            if mandatory:
                msg = "missing auth.ldap.{} in master config".format(key)
                raise SaltInvocationError(msg)
            return False
    return value


def _render_template(param, username):
    """
    Render config template, substituting username where found.
    """
    env = Environment()
    template = env.from_string(param)
    variables = {"username": username}
    return template.render(variables)


class _LDAPConnection:
    """
    Setup an LDAP connection.
    """

    def __init__(
        self,
        uri,
        server,
        port,
        starttls,
        tls,
        no_verify,
        binddn,
        bindpw,
        anonymous,
        accountattributename,
        activedirectory=False,
    ):
        """
        Bind to an LDAP directory using passed credentials.
        """
        self.uri = uri
        self.server = server
        self.port = port
        self.starttls = starttls
        self.tls = tls
        self.binddn = binddn
        self.bindpw = bindpw
        if not HAS_LDAP:
            raise CommandExecutionError(
                "LDAP connection could not be made, the python-ldap module is "
                "not installed. Install python-ldap to use LDAP external auth."
            )
        if self.starttls and self.tls:
            raise CommandExecutionError(
                "Cannot bind with both starttls and tls enabled."
                "Please enable only one of the protocols"
            )

        schema = "ldaps" if tls else "ldap"
        if self.uri == "":
            self.uri = "{}://{}:{}".format(schema, self.server, self.port)

        try:
            if no_verify:
                ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

            self.ldap = ldap.initialize("{}".format(self.uri))
            self.ldap.protocol_version = 3  # ldap.VERSION3
            self.ldap.set_option(ldap.OPT_REFERRALS, 0)  # Needed for AD

            if not anonymous:
                if not self.bindpw:
                    raise CommandExecutionError(
                        "LDAP bind password is not set: password cannot be empty if auth.ldap.anonymous is False"
                    )
                if self.starttls:
                    self.ldap.start_tls_s()
                self.ldap.simple_bind_s(self.binddn, self.bindpw)
        except Exception as ldap_error:  # pylint: disable=broad-except
            raise CommandExecutionError(
                "Failed to bind to LDAP server {} as {}: {}".format(
                    self.uri, self.binddn, ldap_error
                )
            )


def _bind_for_search(anonymous=False, opts=None):
    """
    Bind with binddn and bindpw only for searching LDAP
    :param anonymous: Try binding anonymously
    :param opts: Pass in when __opts__ is not available
    :return: LDAPConnection object
    """
    # Get config params; create connection dictionary
    connargs = {}
    # config params (auth.ldap.*)
    params = {
        "mandatory": [
            "uri",
            "server",
            "port",
            "starttls",
            "tls",
            "no_verify",
            "anonymous",
            "accountattributename",
            "activedirectory",
        ],
        "additional": [
            "binddn",
            "bindpw",
            "filter",
            "groupclass",
            "auth_by_group_membership_only",
        ],
    }

    paramvalues = {}

    for param in params["mandatory"]:
        paramvalues[param] = _config(param, opts=opts)

    for param in params["additional"]:
        paramvalues[param] = _config(param, mandatory=False, opts=opts)

    paramvalues["anonymous"] = anonymous

    # Only add binddn/bindpw to the connargs when they're set, as they're not
    # mandatory for initializing the LDAP object, but if they're provided
    # initially, a bind attempt will be done during the initialization to
    # validate them
    if paramvalues["binddn"]:
        connargs["binddn"] = paramvalues["binddn"]
        if paramvalues["bindpw"]:
            params["mandatory"].append("bindpw")

    for name in params["mandatory"]:
        connargs[name] = paramvalues[name]

    if not paramvalues["anonymous"]:
        if paramvalues["binddn"] and paramvalues["bindpw"]:
            # search for the user's DN to be used for the actual authentication
            return _LDAPConnection(**connargs).ldap


def _bind(username, password, anonymous=False, opts=None):
    """
    Authenticate via an LDAP bind
    """
    # Get config params; create connection dictionary
    basedn = _config("basedn", opts=opts)
    scope = _config("scope", opts=opts)
    connargs = {}
    # config params (auth.ldap.*)
    params = {
        "mandatory": [
            "uri",
            "server",
            "port",
            "starttls",
            "tls",
            "no_verify",
            "anonymous",
            "accountattributename",
            "activedirectory",
        ],
        "additional": [
            "binddn",
            "bindpw",
            "filter",
            "groupclass",
            "auth_by_group_membership_only",
        ],
    }

    paramvalues = {}

    for param in params["mandatory"]:
        paramvalues[param] = _config(param, opts=opts)

    for param in params["additional"]:
        paramvalues[param] = _config(param, mandatory=False, opts=opts)

    paramvalues["anonymous"] = anonymous
    if paramvalues["binddn"]:
        # the binddn can also be composited, e.g.
        #   - {{ username }}@domain.com
        #   - cn={{ username }},ou=users,dc=company,dc=tld
        # so make sure to render it first before using it
        paramvalues["binddn"] = _render_template(paramvalues["binddn"], username)
        paramvalues["binddn"] = ldap.filter.escape_filter_chars(paramvalues["binddn"])

    if paramvalues["filter"]:
        escaped_username = ldap.filter.escape_filter_chars(username)
        paramvalues["filter"] = _render_template(
            paramvalues["filter"], escaped_username
        )

    # Only add binddn/bindpw to the connargs when they're set, as they're not
    # mandatory for initializing the LDAP object, but if they're provided
    # initially, a bind attempt will be done during the initialization to
    # validate them
    if paramvalues["binddn"]:
        connargs["binddn"] = paramvalues["binddn"]
        if paramvalues["bindpw"]:
            params["mandatory"].append("bindpw")

    for name in params["mandatory"]:
        connargs[name] = paramvalues[name]

    if not paramvalues["anonymous"]:
        if paramvalues["binddn"] and paramvalues["bindpw"]:
            # search for the user's DN to be used for the actual authentication
            _ldap = _LDAPConnection(**connargs).ldap
            log.debug(
                "Running LDAP user dn search with filter:%s, dn:%s, scope:%s",
                paramvalues["filter"],
                basedn,
                scope,
            )
            result = _ldap.search_s(basedn, int(scope), paramvalues["filter"])
            if not result:
                log.warning("Unable to find user %s", username)
                return False
            elif len(result) > 1:
                # Active Directory returns something odd.  Though we do not
                # chase referrals (ldap.set_option(ldap.OPT_REFERRALS, 0) above)
                # it still appears to return several entries for other potential
                # sources for a match.  All these sources have None for the
                # CN (ldap array return items are tuples: (cn, ldap entry))
                # But the actual CNs are at the front of the list.
                # So with some list comprehension magic, extract the first tuple
                # entry from all the results, create a list from those,
                # and count the ones that are not None.  If that total is more than one
                # we need to error out because the ldap filter isn't narrow enough.
                cns = [tup[0] for tup in result]
                total_not_none = sum(1 for c in cns if c is not None)
                if total_not_none > 1:
                    log.error(
                        "LDAP lookup found multiple results for user %s", username
                    )
                    return False
                elif total_not_none == 0:
                    log.error(
                        "LDAP lookup--unable to find CN matching user %s", username
                    )
                    return False

            connargs["binddn"] = result[0][0]
        if paramvalues["binddn"] and not paramvalues["bindpw"]:
            connargs["binddn"] = paramvalues["binddn"]
    elif paramvalues["binddn"] and not paramvalues["bindpw"]:
        connargs["binddn"] = paramvalues["binddn"]

    # Update connection dictionary with the user's password
    connargs["bindpw"] = password

    # Attempt bind with user dn and password
    if paramvalues["anonymous"]:
        log.debug("Attempting anonymous LDAP bind")
    else:
        log.debug("Attempting LDAP bind with user dn: %s", connargs["binddn"])
    try:
        ldap_conn = _LDAPConnection(**connargs).ldap
    except Exception:  # pylint: disable=broad-except
        connargs.pop("bindpw", None)  # Don't log the password
        log.error("Failed to authenticate user dn via LDAP: %s", connargs)
        log.debug("Error authenticating user dn via LDAP:", exc_info=True)
        return False
    log.debug("Successfully authenticated user dn via LDAP: %s", connargs["binddn"])
    return ldap_conn


def auth(username, password):
    """
    Simple LDAP auth
    """
    if not HAS_LDAP:
        log.error("LDAP authentication requires python-ldap module")
        return False

    bind = None

    # If bind credentials are configured, verify that we receive a valid bind
    if _config("binddn", mandatory=False) and _config("bindpw", mandatory=False):
        search_bind = _bind_for_search(anonymous=_config("anonymous", mandatory=False))

        # If username & password are not None, attempt to verify they are valid
        if search_bind and username and password:
            bind = _bind(
                username,
                password,
                anonymous=_config("auth_by_group_membership_only", mandatory=False)
                and _config("anonymous", mandatory=False),
            )
    else:
        bind = _bind(
            username,
            password,
            anonymous=_config("auth_by_group_membership_only", mandatory=False)
            and _config("anonymous", mandatory=False),
        )

    if bind:
        log.debug("LDAP authentication successful")
        return bind

    log.error("LDAP _bind authentication FAILED")
    return False


def groups(username, **kwargs):
    """
    Authenticate against an LDAP group

    Behavior is highly dependent on if Active Directory is in use.

    AD handles group membership very differently than OpenLDAP.
    See the :ref:`External Authentication <acl-eauth>` documentation for a thorough
    discussion of available parameters for customizing the search.

    OpenLDAP allows you to search for all groups in the directory
    and returns members of those groups.  Then we check against
    the username entered.

    """
    group_list = []

    # If bind credentials are configured, use them instead of user's
    if _config("binddn", mandatory=False) and _config("bindpw", mandatory=False):
        bind = _bind_for_search(anonymous=_config("anonymous", mandatory=False))
    else:
        bind = _bind(
            username,
            kwargs.get("password", ""),
            anonymous=_config("auth_by_group_membership_only", mandatory=False)
            and _config("anonymous", mandatory=False),
        )

    if bind:
        log.debug("ldap bind to determine group membership succeeded!")

        if _config("activedirectory"):
            try:
                get_user_dn_search = "(&({}={})(objectClass={}))".format(
                    _config("accountattributename"), username, _config("persontype")
                )
                user_dn_results = bind.search_s(
                    _config("basedn"),
                    ldap.SCOPE_SUBTREE,
                    get_user_dn_search,
                    ["distinguishedName"],
                )
            except Exception as e:  # pylint: disable=broad-except
                log.error("Exception thrown while looking up user DN in AD: %s", e)
                return group_list
            if not user_dn_results:
                log.error("Could not get distinguished name for user %s", username)
                return group_list
            # LDAP results are always tuples.  First entry in the tuple is the DN
            dn = ldap.filter.escape_filter_chars(user_dn_results[0][0])
            ldap_search_string = "(&(member={})(objectClass={}))".format(
                dn, _config("groupclass")
            )
            log.debug("Running LDAP group membership search: %s", ldap_search_string)
            try:
                search_results = bind.search_s(
                    _config("basedn"),
                    ldap.SCOPE_SUBTREE,
                    ldap_search_string,
                    [
                        salt.utils.stringutils.to_str(_config("accountattributename")),
                        "cn",
                    ],
                )
            except Exception as e:  # pylint: disable=broad-except
                log.error(
                    "Exception thrown while retrieving group membership in AD: %s", e
                )
                return group_list
            for _, entry in search_results:
                if "cn" in entry:
                    group_list.append(salt.utils.stringutils.to_unicode(entry["cn"][0]))
            log.debug("User %s is a member of groups: %s", username, group_list)

        elif _config("freeipa"):
            escaped_username = ldap.filter.escape_filter_chars(username)
            search_base = _config("group_basedn")
            search_string = _render_template(_config("group_filter"), escaped_username)
            search_results = bind.search_s(
                search_base,
                ldap.SCOPE_SUBTREE,
                search_string,
                [
                    salt.utils.stringutils.to_str(_config("accountattributename")),
                    salt.utils.stringutils.to_str(_config("groupattribute")),
                    "cn",
                ],
            )

            for entry, result in search_results:
                for user in itertools.chain(
                    result.get(_config("accountattributename"), []),
                    result.get(_config("groupattribute"), []),
                ):
                    if (
                        username
                        == salt.utils.stringutils.to_unicode(user)
                        .split(",")[0]
                        .split("=")[-1]
                    ):
                        group_list.append(entry.split(",")[0].split("=")[-1])

            log.debug("User %s is a member of groups: %s", username, group_list)

            if not auth(username, kwargs["password"]):
                log.error("LDAP username and password do not match")
                return []
        else:
            if _config("groupou"):
                search_base = "ou={},{}".format(_config("groupou"), _config("basedn"))
            else:
                search_base = "{}".format(_config("basedn"))
            search_string = "(&({}={})(objectClass={}))".format(
                _config("accountattributename"), username, _config("groupclass")
            )
            search_results = bind.search_s(
                search_base,
                ldap.SCOPE_SUBTREE,
                search_string,
                [
                    salt.utils.stringutils.to_str(_config("accountattributename")),
                    "cn",
                    salt.utils.stringutils.to_str(_config("groupattribute")),
                ],
            )
            for _, entry in search_results:
                if username in salt.utils.data.decode(
                    entry[_config("accountattributename")]
                ):
                    group_list.append(salt.utils.stringutils.to_unicode(entry["cn"][0]))
            for user, entry in search_results:
                if (
                    username
                    == salt.utils.stringutils.to_unicode(user)
                    .split(",")[0]
                    .split("=")[-1]
                ):
                    for group in salt.utils.data.decode(
                        entry[_config("groupattribute")]
                    ):
                        group_list.append(
                            salt.utils.stringutils.to_unicode(group)
                            .split(",")[0]
                            .split("=")[-1]
                        )
            log.debug("User %s is a member of groups: %s", username, group_list)

            # Only test user auth on first call for job.
            # 'show_jid' only exists on first payload so we can use that for the conditional.
            if "show_jid" in kwargs and not _bind(
                username,
                kwargs.get("password"),
                anonymous=_config("auth_by_group_membership_only", mandatory=False)
                and _config("anonymous", mandatory=False),
            ):
                log.error("LDAP username and password do not match")
                return []
    else:
        log.error("ldap bind to determine group membership FAILED!")

    return group_list


def __expand_ldap_entries(entries, opts=None):
    """

    :param entries: ldap subtree in external_auth config option
    :param opts: Opts to use when __opts__ not defined
    :return: Dictionary with all allowed operations

    Takes the ldap subtree in the external_auth config option and expands it
    with actual minion names

    webadmins%:  <all users in the AD 'webadmins' group>
      - server1
          - .*
      - ldap(OU=webservers,dc=int,dc=bigcompany,dc=com)
        - test.ping
        - service.restart
      - ldap(OU=Domain Controllers,dc=int,dc=bigcompany,dc=com)
        - allowed_fn_list_attribute^

    This function only gets called if auth.ldap.activedirectory = True
    """
    bind = _bind_for_search(opts=opts)
    acl_tree = []
    for user_or_group_dict in entries:
        if not isinstance(user_or_group_dict, dict):
            acl_tree.append(user_or_group_dict)
            continue
        for minion_or_ou, matchers in user_or_group_dict.items():
            permissions = matchers
            retrieved_minion_ids = []
            if minion_or_ou.startswith("ldap("):
                search_base = minion_or_ou.lstrip("ldap(").rstrip(")")

                search_string = "(objectClass=computer)"
                try:
                    search_results = bind.search_s(
                        search_base, ldap.SCOPE_SUBTREE, search_string, ["cn"]
                    )
                    for ldap_match in search_results:
                        try:
                            minion_id = ldap_match[1]["cn"][0].lower()
                            # Some LDAP/AD trees only have the FQDN of machines
                            # in their computer lists.  auth.minion_stripdomains
                            # lets a user strip off configured domain names
                            # and arrive at the basic minion_id
                            if opts.get("auth.ldap.minion_stripdomains", None):
                                for domain in opts["auth.ldap.minion_stripdomains"]:
                                    if minion_id.endswith(domain):
                                        minion_id = minion_id[: -len(domain)]
                                        break
                            retrieved_minion_ids.append(minion_id)
                        except TypeError:
                            # TypeError here just means that one of the returned
                            # entries didn't match the format we expected
                            # from LDAP.
                            pass

                    for minion_id in retrieved_minion_ids:
                        acl_tree.append({minion_id: permissions})
                    log.trace("Expanded acl_tree is: %s", acl_tree)
                except ldap.NO_SUCH_OBJECT:
                    pass
            else:
                acl_tree.append({minion_or_ou: matchers})

    log.trace("__expand_ldap_entries: %s", acl_tree)
    return acl_tree


def process_acl(auth_list, opts=None):
    """
    Query LDAP, retrieve list of minion_ids from an OU or other search.
    For each minion_id returned from the LDAP search, copy the perms
    matchers into the auth dictionary
    :param auth_list:
    :param opts: __opts__ for when __opts__ is not injected
    :return: Modified auth list.
    """
    ou_names = []
    for item in auth_list:
        if isinstance(item, str):
            continue
        ou_names.extend(
            [
                potential_ou
                for potential_ou in item.keys()
                if potential_ou.startswith("ldap(")
            ]
        )
    if ou_names:
        auth_list = __expand_ldap_entries(auth_list, opts)
    return auth_list



import datetime
import logging
import re

import salt.modules.postgres as postgres
from salt.exceptions import SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import Mock, call, patch
from tests.support.unit import TestCase

test_list_db_csv = (
    "Name,Owner,Encoding,Collate,Ctype,Access privileges,Tablespace\n"
    "template1,postgres,LATIN1,en_US,en_US"
    ',"{=c/postgres,postgres=CTc/postgres}",pg_default\n'
    "template0,postgres,LATIN1,en_US,en_US"
    ',"{=c/postgres,postgres=CTc/postgres}",pg_default\n'
    "postgres,postgres,LATIN1,en_US,en_US,,pg_default\n"
    "test_db,postgres,LATIN1,en_US,en_US,,pg_default"
)

test_list_schema_csv = (
    "name,owner,acl\n"
    'public,postgres,"{postgres=UC/postgres,=UC/postgres}"\n'
    'pg_toast,postgres,""'
)

test_list_language_csv = "Name\ninternal\nc\nsql\nplpgsql\n"

test_privileges_list_table_csv = (
    "name\n"
    '"{baruwatest=arwdDxt/baruwatest,bayestest=arwd/baruwatest,baruwa=a*r*w*d*D*x*t*/baruwatest}"\n'
)

test_privileges_list_group_csv = (
    "rolname,admin_option\nbaruwa,f\nbaruwatest2,t\nbaruwatest,f\n"
)

log = logging.getLogger(__name__)


class PostgresTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        patcher = patch("salt.utils.path.which", Mock(return_value="/usr/bin/pgsql"))
        patcher.start()
        self.addCleanup(patcher.stop)
        return {
            postgres: {
                "__grains__": {"os_family": "Linux"},
                "__salt__": {
                    "config.option": Mock(),
                    "cmd.run_all": Mock(),
                    "file.chown": Mock(),
                    "file.remove": Mock(),
                },
            }
        }

    def test_run_psql(self):
        postgres._run_psql('echo "hi"')
        cmd = postgres.__salt__["cmd.run_all"]

        self.assertEqual("postgres", cmd.call_args[1]["runas"])

    def test_db_alter(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            postgres.db_alter(
                "dbname",
                user="testuser",
                host="testhost",
                port="testport",
                maintenance_db="maint_db",
                password="foo",
                tablespace="testspace",
                owner="otheruser",
                runas="foo",
            )
            postgres._run_psql.assert_has_calls(
                [
                    call(
                        [
                            "/usr/bin/pgsql",
                            "--no-align",
                            "--no-readline",
                            "--no-psqlrc",
                            "--no-password",
                            "--username",
                            "testuser",
                            "--host",
                            "testhost",
                            "--port",
                            "testport",
                            "--dbname",
                            "maint_db",
                            "-c",
                            'ALTER DATABASE "dbname" OWNER TO "otheruser"',
                        ],
                        host="testhost",
                        user="testuser",
                        password="foo",
                        runas="foo",
                        port="testport",
                    ),
                    call(
                        [
                            "/usr/bin/pgsql",
                            "--no-align",
                            "--no-readline",
                            "--no-psqlrc",
                            "--no-password",
                            "--username",
                            "testuser",
                            "--host",
                            "testhost",
                            "--port",
                            "testport",
                            "--dbname",
                            "maint_db",
                            "-c",
                            'ALTER DATABASE "dbname" SET TABLESPACE "testspace"',
                        ],
                        host="testhost",
                        user="testuser",
                        password="foo",
                        runas="foo",
                        port="testport",
                    ),
                ]
            )

    def test_db_alter_owner_recurse(self):
        with patch(
            "salt.modules.postgres.owner_to", Mock(return_value={"retcode": None})
        ):
            postgres.db_alter(
                "dbname",
                user="testuser",
                host="testhost",
                port="testport",
                maintenance_db="maint_db",
                password="foo",
                tablespace="testspace",
                owner="otheruser",
                owner_recurse=True,
                runas="foo",
            )
            postgres.owner_to.assert_called_once_with(
                "dbname",
                "otheruser",
                user="testuser",
                host="testhost",
                port="testport",
                password="foo",
                runas="foo",
            )

    def test_db_create(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            postgres.db_create(
                "dbname",
                user="testuser",
                host="testhost",
                port="testport",
                maintenance_db="maint_db",
                password="foo",
                tablespace="testspace",
                owner="otheruser",
                runas="foo",
            )

            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "maint_db",
                    "-c",
                    'CREATE DATABASE "dbname" WITH TABLESPACE = "testspace" '
                    'OWNER = "otheruser"',
                ],
                host="testhost",
                user="testuser",
                password="foo",
                runas="foo",
                port="testport",
            )

    def test_db_create_empty_string_param(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            postgres.db_create(
                "dbname",
                lc_collate="",
                encoding="utf8",
                user="testuser",
                host="testhost",
                port=1234,
                maintenance_db="maint_db",
                password="foo",
            )

            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "1234",
                    "--dbname",
                    "maint_db",
                    "-c",
                    "CREATE DATABASE \"dbname\" WITH ENCODING = 'utf8' LC_COLLATE = ''",
                ],
                host="testhost",
                password="foo",
                port=1234,
                runas=None,
                user="testuser",
            )

    def test_db_create_with_trivial_sql_injection(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            self.assertRaises(
                SaltInvocationError,
                postgres.db_create,
                "dbname",
                lc_collate="foo' ENCODING='utf8",
            )

    def test_db_exists(self):
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_list_db_csv}),
        ):
            ret = postgres.db_exists(
                "test_db",
                user="testuser",
                host="testhost",
                port="testport",
                maintenance_db="maint_db",
                password="foo",
                runas="foo",
            )
            self.assertTrue(ret)

    def test_db_list(self):
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_list_db_csv}),
        ):
            ret = postgres.db_list(
                user="testuser",
                host="testhost",
                port="testport",
                maintenance_db="maint_db",
                password="foo",
                runas="foo",
            )
            self.assertDictEqual(
                ret,
                {
                    "test_db": {
                        "Encoding": "LATIN1",
                        "Ctype": "en_US",
                        "Tablespace": "pg_default",
                        "Collate": "en_US",
                        "Owner": "postgres",
                        "Access privileges": "",
                    },
                    "template1": {
                        "Encoding": "LATIN1",
                        "Ctype": "en_US",
                        "Tablespace": "pg_default",
                        "Collate": "en_US",
                        "Owner": "postgres",
                        "Access privileges": "{=c/postgres,postgres=CTc/postgres}",
                    },
                    "template0": {
                        "Encoding": "LATIN1",
                        "Ctype": "en_US",
                        "Tablespace": "pg_default",
                        "Collate": "en_US",
                        "Owner": "postgres",
                        "Access privileges": "{=c/postgres,postgres=CTc/postgres}",
                    },
                    "postgres": {
                        "Encoding": "LATIN1",
                        "Ctype": "en_US",
                        "Tablespace": "pg_default",
                        "Collate": "en_US",
                        "Owner": "postgres",
                        "Access privileges": "",
                    },
                },
            )

    def test_db_remove(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            postgres.db_remove(
                "test_db",
                user="testuser",
                host="testhost",
                port="testport",
                maintenance_db="maint_db",
                password="foo",
                runas="foo",
            )

            calls = (
                call(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        'REVOKE CONNECT ON DATABASE "test_db" FROM public;',
                    ],
                    host="testhost",
                    password="foo",
                    port="testport",
                    runas="foo",
                    user="testuser",
                ),
                call(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        "SELECT pid, pg_terminate_backend(pid) FROM pg_stat_activity"
                        " WHERE datname = 'test_db' AND pid <> pg_backend_pid();",
                    ],
                    host="testhost",
                    password="foo",
                    port="testport",
                    runas="foo",
                    user="testuser",
                ),
                call(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        'DROP DATABASE "test_db";',
                    ],
                    host="testhost",
                    password="foo",
                    port="testport",
                    runas="foo",
                    user="testuser",
                ),
            )

            postgres._run_psql.assert_has_calls(calls, any_order=True)

    def test_group_create(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.user_exists", Mock(return_value=False)):
                postgres.group_create(
                    "testgroup",
                    user="testuser",
                    host="testhost",
                    port="testport",
                    maintenance_db="maint_db",
                    password="foo",
                    createdb=False,
                    encrypted=False,
                    superuser=False,
                    replication=False,
                    rolepassword="testrolepass",
                    groups="testgroup",
                    runas="foo",
                )
                # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
                # The first 14 elements of this list are initial args used in all (or
                # virtually all) commands run through _run_psql(), so the actual SQL
                # query will be in the 15th argument.
                self.assertTrue(
                    postgres._run_psql.call_args[0][0][14].startswith("CREATE ROLE")
                )

    def test_group_remove(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.user_exists", Mock(return_value=True)):
                postgres.group_remove(
                    "testgroup",
                    user="testuser",
                    host="testhost",
                    port="testport",
                    maintenance_db="maint_db",
                    password="foo",
                    runas="foo",
                )
                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        'DROP ROLE "testgroup"',
                    ],
                    host="testhost",
                    user="testuser",
                    password="foo",
                    runas="foo",
                    port="testport",
                )

    def test_group_update(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.role_get",
                Mock(return_value={"superuser": False}),
            ):
                postgres.group_update(
                    "testgroup",
                    user='"testuser"',
                    host="testhost",
                    port="testport",
                    maintenance_db="maint_db",
                    password="foo",
                    createdb=False,
                    encrypted=False,
                    replication=False,
                    rolepassword="test_role_pass",
                    groups="testgroup",
                    runas="foo",
                )
                # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
                # The first 14 elements of this list are initial args used in all (or
                # virtually all) commands run through _run_psql(), so the actual SQL
                # query will be in the 15th argument.
                self.assertTrue(
                    re.match(
                        'ALTER.* "testgroup" .* UNENCRYPTED PASSWORD',
                        postgres._run_psql.call_args[0][0][14],
                    )
                )

    def test_user_create(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.user_exists", Mock(return_value=False)):
                postgres.user_create(
                    "testuser",
                    user="testuser",
                    host="testhost",
                    port="testport",
                    maintenance_db="maint_test",
                    password="test_pass",
                    login=True,
                    createdb=False,
                    createroles=False,
                    encrypted=False,
                    superuser=False,
                    replication=False,
                    rolepassword="test_role_pass",
                    valid_until="2042-07-01",
                    groups="test_groups",
                    runas="foo",
                )
                # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
                # The first 14 elements of this list are initial args used in all (or
                # virtually all) commands run through _run_psql(), so the actual SQL
                # query will be in the 15th argument.
                call = postgres._run_psql.call_args[0][0][14]
                self.assertTrue(re.match('CREATE ROLE "testuser"', call))
                for i in (
                    "INHERIT",
                    "NOCREATEDB",
                    "NOCREATEROLE",
                    "NOSUPERUSER",
                    "NOREPLICATION",
                    "LOGIN",
                    "UNENCRYPTED",
                    "PASSWORD",
                    "VALID UNTIL",
                ):
                    self.assertTrue(i in call, "{} not in {}".format(i, call))

    def test_user_exists(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.version", Mock(return_value="9.1")):
                with patch(
                    "salt.modules.postgres.psql_query",
                    Mock(
                        return_value=[
                            {
                                "name": "test_user",
                                "superuser": "t",
                                "inherits privileges": "t",
                                "can create roles": "t",
                                "can create databases": "t",
                                "can update system catalogs": "t",
                                "can login": "t",
                                "replication": None,
                                "password": "test_password",
                                "connections": "-1",
                                "groups": "",
                                "expiry time": "",
                                "defaults variables": None,
                            }
                        ]
                    ),
                ):
                    ret = postgres.user_exists(
                        "test_user",
                        user="test_user",
                        host="test_host",
                        port="test_port",
                        maintenance_db="maint_db",
                        password="test_password",
                        runas="foo",
                    )
                    self.assertTrue(ret)

    def test_user_list(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.version", Mock(return_value="9.1")):
                with patch(
                    "salt.modules.postgres.psql_query",
                    Mock(
                        return_value=[
                            {
                                "name": "test_user",
                                "superuser": "t",
                                "inherits privileges": "t",
                                "can create roles": "t",
                                "can create databases": "t",
                                "can update system catalogs": "t",
                                "can login": "t",
                                "replication": None,
                                "connections": "-1",
                                "groups": "",
                                "expiry time": "2017-08-16 08:57:46",
                                "defaults variables": None,
                            }
                        ]
                    ),
                ):
                    ret = postgres.user_list(
                        "test_user",
                        host="test_host",
                        port="test_port",
                        maintenance_db="maint_db",
                        password="test_password",
                        runas="foo",
                    )

                    self.assertDictEqual(
                        ret,
                        {
                            "test_user": {
                                "superuser": True,
                                "defaults variables": None,
                                "can create databases": True,
                                "can create roles": True,
                                "connections": None,
                                "replication": None,
                                "expiry time": datetime.datetime(
                                    2017, 8, 16, 8, 57, 46
                                ),
                                "can login": True,
                                "can update system catalogs": True,
                                "groups": [],
                                "inherits privileges": True,
                            }
                        },
                    )

    def test_user_remove(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.version", Mock(return_value="9.1")):
                with patch(
                    "salt.modules.postgres.user_exists", Mock(return_value=True)
                ):
                    postgres.user_remove(
                        "testuser",
                        user="testuser",
                        host="testhost",
                        port="testport",
                        maintenance_db="maint_db",
                        password="testpassword",
                        runas="foo",
                    )
                    postgres._run_psql.assert_called_once_with(
                        [
                            "/usr/bin/pgsql",
                            "--no-align",
                            "--no-readline",
                            "--no-psqlrc",
                            "--no-password",
                            "--username",
                            "testuser",
                            "--host",
                            "testhost",
                            "--port",
                            "testport",
                            "--dbname",
                            "maint_db",
                            "-c",
                            'DROP ROLE "testuser"',
                        ],
                        host="testhost",
                        port="testport",
                        user="testuser",
                        password="testpassword",
                        runas="foo",
                    )

    def test_user_update(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.role_get",
                Mock(return_value={"superuser": False}),
            ):
                postgres.user_update(
                    "test_username",
                    user="test_user",
                    host="test_host",
                    port="test_port",
                    maintenance_db="test_maint",
                    password="test_pass",
                    createdb=False,
                    createroles=False,
                    encrypted=False,
                    inherit=True,
                    login=True,
                    replication=False,
                    rolepassword="test_role_pass",
                    valid_until="2017-07-01",
                    groups="test_groups",
                    runas="foo",
                )
                # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
                # The first 14 elements of this list are initial args used in all (or
                # virtually all) commands run through _run_psql(), so the actual SQL
                # query will be in the 15th argument.
                self.assertTrue(
                    re.match(
                        'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
                        "NOCREATEROLE NOREPLICATION LOGIN "
                        "UNENCRYPTED PASSWORD ['\"]{0,5}test_role_pass['\"]{0,5} "
                        "VALID UNTIL '2017-07-01';"
                        ' GRANT "test_groups" TO "test_username"',
                        postgres._run_psql.call_args[0][0][14],
                    )
                )

    def test_user_update2(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.role_get",
                Mock(return_value={"superuser": False}),
            ):
                postgres.user_update(
                    "test_username",
                    user="test_user",
                    host="test_host",
                    port="test_port",
                    maintenance_db="test_maint",
                    password="test_pass",
                    createdb=False,
                    createroles=True,
                    encrypted=False,
                    inherit=True,
                    login=True,
                    replication=False,
                    groups="test_groups",
                    runas="foo",
                )
                # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
                # The first 14 elements of this list are initial args used in all (or
                # virtually all) commands run through _run_psql(), so the actual SQL
                # query will be in the 15th argument.
                self.assertTrue(
                    re.match(
                        'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
                        "CREATEROLE NOREPLICATION LOGIN;"
                        ' GRANT "test_groups" TO "test_username"',
                        postgres._run_psql.call_args[0][0][14],
                    )
                )

    def test_user_update3(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.role_get",
                Mock(return_value={"superuser": False}),
            ):
                postgres.user_update(
                    "test_username",
                    user="test_user",
                    host="test_host",
                    port="test_port",
                    maintenance_db="test_maint",
                    password="test_pass",
                    createdb=False,
                    createroles=True,
                    encrypted=False,
                    inherit=True,
                    login=True,
                    rolepassword=False,
                    replication=False,
                    groups="test_groups",
                    runas="foo",
                )
                # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
                # The first 14 elements of this list are initial args used in all (or
                # virtually all) commands run through _run_psql(), so the actual SQL
                # query will be in the 15th argument.
                self.assertTrue(
                    re.match(
                        'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
                        "CREATEROLE NOREPLICATION LOGIN NOPASSWORD;"
                        ' GRANT "test_groups" TO "test_username"',
                        postgres._run_psql.call_args[0][0][14],
                    )
                )

    def test_user_update_encrypted_passwd(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.role_get",
                Mock(return_value={"superuser": False}),
            ):
                postgres.user_update(
                    "test_username",
                    user="test_user",
                    host="test_host",
                    port="test_port",
                    maintenance_db="test_maint",
                    password="test_pass",
                    createdb=False,
                    createroles=True,
                    encrypted=True,
                    inherit=True,
                    login=True,
                    rolepassword="foobar",
                    replication=False,
                    groups="test_groups",
                    runas="foo",
                )
                # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
                # The first 14 elements of this list are initial args used in all (or
                # virtually all) commands run through _run_psql(), so the actual SQL
                # query will be in the 15th argument.
                self.assertTrue(
                    re.match(
                        'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
                        "CREATEROLE NOREPLICATION LOGIN "
                        "ENCRYPTED PASSWORD "
                        "['\"]{0,5}md531c27e68d3771c392b52102c01be1da1['\"]{0,5}"
                        '; GRANT "test_groups" TO "test_username"',
                        postgres._run_psql.call_args[0][0][14],
                    )
                )

    def test_version(self):
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": "9.1.9"}),
        ):
            postgres.version(
                user="test_user",
                host="test_host",
                port="test_port",
                maintenance_db="test_maint",
                password="test_pass",
                runas="foo",
            )
            # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
            # The first 14 elements of this list are initial args used in all (or
            # virtually all) commands run through _run_psql(), so the actual SQL
            # query will be in the 15th argument.
            self.assertTrue(
                re.match(
                    "SELECT setting FROM pg_catalog.pg_settings",
                    postgres._run_psql.call_args[0][0][14],
                )
            )

    def test_installed_extensions(self):
        with patch(
            "salt.modules.postgres.psql_query",
            Mock(return_value=[{"extname": "foo", "extversion": "1"}]),
        ):
            exts = postgres.installed_extensions()
            self.assertEqual(exts, {"foo": {"extversion": "1", "extname": "foo"}})

    def test_available_extensions(self):
        with patch(
            "salt.modules.postgres.psql_query",
            Mock(return_value=[{"name": "foo", "default_version": "1"}]),
        ):
            exts = postgres.available_extensions()
            self.assertEqual(exts, {"foo": {"default_version": "1", "name": "foo"}})

    def test_drop_extension2(self):
        with patch(
            "salt.modules.postgres.installed_extensions", Mock(side_effect=[{}, {}])
        ):
            with patch(
                "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
            ):
                with patch(
                    "salt.modules.postgres.available_extensions",
                    Mock(return_value={"foo": {"default_version": "1", "name": "foo"}}),
                ):
                    self.assertEqual(postgres.drop_extension("foo"), True)

    def test_drop_extension3(self):
        with patch(
            "salt.modules.postgres.installed_extensions",
            Mock(side_effect=[{"foo": {"extversion": "1", "extname": "foo"}}, {}]),
        ):
            with patch(
                "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
            ):
                with patch(
                    "salt.modules.postgres.available_extensions",
                    Mock(return_value={"foo": {"default_version": "1", "name": "foo"}}),
                ):
                    self.assertEqual(postgres.drop_extension("foo"), True)

    def test_drop_extension1(self):
        with patch(
            "salt.modules.postgres.installed_extensions",
            Mock(
                side_effect=[
                    {"foo": {"extversion": "1", "extname": "foo"}},
                    {"foo": {"extversion": "1", "extname": "foo"}},
                ]
            ),
        ):
            with patch(
                "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
            ):
                with patch(
                    "salt.modules.postgres.available_extensions",
                    Mock(return_value={"foo": {"default_version": "1", "name": "foo"}}),
                ):
                    self.assertEqual(postgres.drop_extension("foo"), False)

    def test_create_mtdata(self):
        with patch(
            "salt.modules.postgres.installed_extensions",
            Mock(
                return_value={
                    "foo": {
                        "extversion": "0.8",
                        "extrelocatable": "t",
                        "schema_name": "foo",
                        "extname": "foo",
                    }
                },
            ),
        ):
            with patch(
                "salt.modules.postgres.available_extensions",
                Mock(return_value={"foo": {"default_version": "1.4", "name": "foo"}}),
            ):
                ret = postgres.create_metadata("foo", schema="bar", ext_version="1.4")
                self.assertTrue(postgres._EXTENSION_INSTALLED in ret)
                self.assertTrue(postgres._EXTENSION_TO_UPGRADE in ret)
                self.assertTrue(postgres._EXTENSION_TO_MOVE in ret)
                ret = postgres.create_metadata("foo", schema="foo", ext_version="0.4")
                self.assertTrue(postgres._EXTENSION_INSTALLED in ret)
                self.assertFalse(postgres._EXTENSION_TO_UPGRADE in ret)
                self.assertFalse(postgres._EXTENSION_TO_MOVE in ret)
                ret = postgres.create_metadata("foo")
                self.assertTrue(postgres._EXTENSION_INSTALLED in ret)
                self.assertFalse(postgres._EXTENSION_TO_UPGRADE in ret)
                self.assertFalse(postgres._EXTENSION_TO_MOVE in ret)
                ret = postgres.create_metadata("foobar")
                self.assertTrue(postgres._EXTENSION_NOT_INSTALLED in ret)
                self.assertFalse(postgres._EXTENSION_INSTALLED in ret)
                self.assertFalse(postgres._EXTENSION_TO_UPGRADE in ret)
                self.assertFalse(postgres._EXTENSION_TO_MOVE in ret)

    def test_create_extension_newerthan(self):
        """
        scenario of creating upgrading extensions with possible schema and
        version specifications
        """
        with patch(
            "salt.modules.postgres.create_metadata",
            Mock(
                side_effect=[
                    # create succeeded
                    [postgres._EXTENSION_NOT_INSTALLED],
                    [postgres._EXTENSION_INSTALLED],
                    [postgres._EXTENSION_NOT_INSTALLED],
                    [postgres._EXTENSION_INSTALLED],
                    # create failed
                    [postgres._EXTENSION_NOT_INSTALLED],
                    [postgres._EXTENSION_NOT_INSTALLED],
                    # move+upgrade succeeded
                    [
                        postgres._EXTENSION_TO_MOVE,
                        postgres._EXTENSION_TO_UPGRADE,
                        postgres._EXTENSION_INSTALLED,
                    ],
                    [postgres._EXTENSION_INSTALLED],
                    # move succeeded
                    [postgres._EXTENSION_TO_MOVE, postgres._EXTENSION_INSTALLED],
                    [postgres._EXTENSION_INSTALLED],
                    # upgrade succeeded
                    [postgres._EXTENSION_TO_UPGRADE, postgres._EXTENSION_INSTALLED],
                    [postgres._EXTENSION_INSTALLED],
                    # upgrade failed
                    [postgres._EXTENSION_TO_UPGRADE, postgres._EXTENSION_INSTALLED],
                    [postgres._EXTENSION_TO_UPGRADE, postgres._EXTENSION_INSTALLED],
                    # move failed
                    [postgres._EXTENSION_TO_MOVE, postgres._EXTENSION_INSTALLED],
                    [postgres._EXTENSION_TO_MOVE, postgres._EXTENSION_INSTALLED],
                ]
            ),
        ):
            with patch(
                "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
            ):
                with patch(
                    "salt.modules.postgres.available_extensions",
                    Mock(
                        return_value={"foo": {"default_version": "1.4", "name": "foo"}}
                    ),
                ):
                    self.assertTrue(postgres.create_extension("foo"))
                    self.assertTrue(
                        re.match(
                            'CREATE EXTENSION IF NOT EXISTS "foo" ;',
                            postgres._psql_prepare_and_run.call_args[0][0][1],
                        )
                    )
                    self.assertTrue(
                        postgres.create_extension(
                            "foo", schema="a", ext_version="b", from_version="c"
                        )
                    )
                    self.assertTrue(
                        re.match(
                            'CREATE EXTENSION IF NOT EXISTS "foo" '
                            'WITH SCHEMA "a" VERSION b FROM c ;',
                            postgres._psql_prepare_and_run.call_args[0][0][1],
                        )
                    )
                    self.assertFalse(postgres.create_extension("foo"))
                    ret = postgres.create_extension("foo", ext_version="a", schema="b")
                    self.assertTrue(ret)
                    self.assertTrue(
                        re.match(
                            'ALTER EXTENSION "foo" SET SCHEMA "b";'
                            ' ALTER EXTENSION "foo" UPDATE TO a;',
                            postgres._psql_prepare_and_run.call_args[0][0][1],
                        )
                    )
                    ret = postgres.create_extension("foo", ext_version="a", schema="b")
                    self.assertTrue(ret)
                    self.assertTrue(
                        re.match(
                            'ALTER EXTENSION "foo" SET SCHEMA "b";',
                            postgres._psql_prepare_and_run.call_args[0][0][1],
                        )
                    )
                    ret = postgres.create_extension("foo", ext_version="a", schema="b")
                    self.assertTrue(ret)
                    self.assertTrue(
                        re.match(
                            'ALTER EXTENSION "foo" UPDATE TO a;',
                            postgres._psql_prepare_and_run.call_args[0][0][1],
                        )
                    )
                    self.assertFalse(
                        postgres.create_extension("foo", ext_version="a", schema="b")
                    )
                    self.assertFalse(
                        postgres.create_extension("foo", ext_version="a", schema="b")
                    )

    def test_encrypt_passwords(self):
        self.assertEqual(postgres._maybe_encrypt_password("foo", "bar", False), "bar")
        self.assertEqual(
            postgres._maybe_encrypt_password("foo", "bar", True),
            "md596948aad3fcae80c08a35c9b5958cd89",
        )

    def test_schema_list(self):
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_list_schema_csv}),
        ):
            ret = postgres.schema_list(
                "maint_db",
                db_user="testuser",
                db_host="testhost",
                db_port="testport",
                db_password="foo",
            )
            self.assertDictEqual(
                ret,
                {
                    "public": {
                        "acl": "{postgres=UC/postgres,=UC/postgres}",
                        "owner": "postgres",
                    },
                    "pg_toast": {"acl": "", "owner": "postgres"},
                },
            )

    def test_schema_exists(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.psql_query",
                Mock(
                    return_value=[
                        {
                            "name": "public",
                            "acl": "{postgres=UC/postgres,=UC/postgres}",
                            "owner": "postgres",
                        }
                    ]
                ),
            ):
                ret = postgres.schema_exists("template1", "public")
                self.assertTrue(ret)

    def test_schema_get(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.psql_query",
                Mock(
                    return_value=[
                        {
                            "name": "public",
                            "acl": "{postgres=UC/postgres,=UC/postgres}",
                            "owner": "postgres",
                        }
                    ]
                ),
            ):
                ret = postgres.schema_get("template1", "public")
                self.assertTrue(ret)

    def test_schema_get_again(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.psql_query",
                Mock(
                    return_value=[
                        {
                            "name": "public",
                            "acl": "{postgres=UC/postgres,=UC/postgres}",
                            "owner": "postgres",
                        }
                    ]
                ),
            ):
                ret = postgres.schema_get("template1", "pg_toast")
                self.assertFalse(ret)

    def test_schema_create(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.schema_exists", Mock(return_value=False)):
                postgres.schema_create(
                    "maint_db",
                    "testschema",
                    user="user",
                    db_host="testhost",
                    db_port="testport",
                    db_user="testuser",
                    db_password="testpassword",
                )
                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        'CREATE SCHEMA "testschema"',
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_schema_create2(self):
        with patch("salt.modules.postgres.schema_exists", Mock(return_value=True)):
            ret = postgres.schema_create(
                "test_db",
                "test_schema",
                user="user",
                db_host="test_host",
                db_port="test_port",
                db_user="test_user",
                db_password="test_password",
            )
            self.assertFalse(ret)

    def test_schema_remove(self):
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.schema_exists", Mock(return_value=True)):
                postgres.schema_remove(
                    "maint_db",
                    "testschema",
                    user="user",
                    db_host="testhost",
                    db_port="testport",
                    db_user="testuser",
                    db_password="testpassword",
                )
                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        'DROP SCHEMA "testschema"',
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_schema_remove2(self):
        with patch("salt.modules.postgres.schema_exists", Mock(return_value=False)):
            ret = postgres.schema_remove(
                "test_db",
                "test_schema",
                user="user",
                db_host="test_host",
                db_port="test_port",
                db_user="test_user",
                db_password="test_password",
            )
            self.assertFalse(ret)

    def test_language_list(self):
        """
        Test language listing
        """
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_list_language_csv}),
        ):
            ret = postgres.language_list(
                "testdb",
                user="testuser",
                host="testhost",
                port="testport",
                password="foo",
            )
            self.assertDictEqual(
                ret,
                {"c": "c", "internal": "internal", "plpgsql": "plpgsql", "sql": "sql"},
            )

    def test_language_exists(self):
        """
        Test language existence check
        """
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.psql_query",
                Mock(
                    return_value=[
                        {"Name": "internal"},
                        {"Name": "c"},
                        {"Name": "sql"},
                        {"Name": "plpgsql"},
                    ]
                ),
            ):
                with patch(
                    "salt.modules.postgres.language_exists", Mock(return_value=True)
                ):
                    ret = postgres.language_exists("sql", "testdb")
                    self.assertTrue(ret)

    def test_language_create(self):
        """
        Test language creation - does not exist in db
        """
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.language_exists", Mock(return_value=False)
            ):
                postgres.language_create(
                    "plpythonu",
                    "testdb",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )
                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "testdb",
                        "-c",
                        "CREATE LANGUAGE plpythonu",
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_language_create_exists(self):
        """
        Test language creation - already exists in db
        """
        with patch("salt.modules.postgres.language_exists", Mock(return_value=True)):
            ret = postgres.language_create(
                "plpythonu",
                "testdb",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )
            self.assertFalse(ret)

    def test_language_remove(self):
        """
        Test language removal - exists in db
        """
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.language_exists", Mock(return_value=True)
            ):
                postgres.language_remove(
                    "plpgsql",
                    "testdb",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )
                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "testdb",
                        "-c",
                        "DROP LANGUAGE plpgsql",
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_language_remove_non_exist(self):
        """
        Test language removal - does not exist in db
        """
        with patch("salt.modules.postgres.language_exists", Mock(return_value=False)):
            ret = postgres.language_remove(
                "plpgsql",
                "testdb",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )
            self.assertFalse(ret)

    def test_privileges_list_table(self):
        """
        Test privilege listing on a table
        """
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_privileges_list_table_csv}),
        ):
            ret = postgres.privileges_list(
                "awl",
                "table",
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )
            expected = {
                "bayestest": {
                    "INSERT": False,
                    "UPDATE": False,
                    "SELECT": False,
                    "DELETE": False,
                },
                "baruwa": {
                    "INSERT": True,
                    "TRUNCATE": True,
                    "UPDATE": True,
                    "TRIGGER": True,
                    "REFERENCES": True,
                    "SELECT": True,
                    "DELETE": True,
                },
                "baruwatest": {
                    "INSERT": False,
                    "TRUNCATE": False,
                    "UPDATE": False,
                    "TRIGGER": False,
                    "REFERENCES": False,
                    "SELECT": False,
                    "DELETE": False,
                },
            }

            self.assertDictEqual(ret, expected)

            query = (
                "COPY (SELECT relacl AS name FROM pg_catalog.pg_class c "
                "JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace "
                "WHERE nspname = 'public' AND relname = 'awl' AND relkind in ('r', 'v') "
                "ORDER BY relname) TO STDOUT WITH CSV HEADER"
            )

            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "db_name",
                    "-v",
                    "datestyle=ISO,MDY",
                    "-c",
                    query,
                ],
                host="testhost",
                port="testport",
                password="testpassword",
                user="testuser",
                runas="user",
            )

    def test_privileges_list_group(self):
        """
        Test privilege listing on a group
        """
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_privileges_list_group_csv}),
        ):
            ret = postgres.privileges_list(
                "admin",
                "group",
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )
            expected = {
                "baruwa": False,
                "baruwatest": False,
                "baruwatest2": True,
            }

            self.assertDictEqual(ret, expected)

            query = (
                "COPY (SELECT rolname, admin_option "
                "FROM pg_catalog.pg_auth_members m JOIN pg_catalog.pg_roles r "
                "ON m.member=r.oid WHERE m.roleid IN (SELECT oid FROM "
                "pg_catalog.pg_roles WHERE rolname='admin') ORDER BY rolname) "
                "TO STDOUT WITH CSV HEADER"
            )

            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "db_name",
                    "-v",
                    "datestyle=ISO,MDY",
                    "-c",
                    query,
                ],
                host="testhost",
                port="testport",
                password="testpassword",
                user="testuser",
                runas="user",
            )

    def test_has_privileges_on_table(self):
        """
        Test privilege checks on table
        """
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_privileges_list_table_csv}),
        ):
            ret = postgres.has_privileges(
                "baruwa",
                "awl",
                "table",
                "SELECT,INSERT",
                grant_option=True,
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertTrue(ret)

            ret = postgres.has_privileges(
                "baruwa",
                "awl",
                "table",
                "ALL",
                grant_option=True,
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertTrue(ret)

            ret = postgres.has_privileges(
                "baruwa",
                "awl",
                "table",
                "ALL",
                grant_option=False,
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertTrue(ret)

            ret = postgres.has_privileges(
                "bayestest",
                "awl",
                "table",
                "SELECT,INSERT,TRUNCATE",
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertFalse(ret)

            ret = postgres.has_privileges(
                "bayestest",
                "awl",
                "table",
                "SELECT,INSERT",
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertTrue(ret)

    def test_has_privileges_on_group(self):
        """
        Test privilege checks on group
        """
        with patch(
            "salt.modules.postgres._run_psql",
            Mock(return_value={"retcode": 0, "stdout": test_privileges_list_group_csv}),
        ):
            ret = postgres.has_privileges(
                "baruwa",
                "admin",
                "group",
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertTrue(ret)

            ret = postgres.has_privileges(
                "baruwa",
                "admin",
                "group",
                grant_option=True,
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertFalse(ret)

            ret = postgres.has_privileges(
                "tony",
                "admin",
                "group",
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            self.assertFalse(ret)

    def test_privileges_grant_table(self):
        """
        Test granting privileges on table
        """
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.has_privileges", Mock(return_value=False)
            ):
                ret = postgres.privileges_grant(
                    "baruwa",
                    "awl",
                    "table",
                    "ALL",
                    grant_option=True,
                    maintenance_db="db_name",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )

                query = 'GRANT ALL ON TABLE public."awl" TO "baruwa" WITH GRANT OPTION'

                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "db_name",
                        "-c",
                        query,
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.has_privileges", Mock(return_value=False)
            ):
                ret = postgres.privileges_grant(
                    "baruwa",
                    "awl",
                    "table",
                    "ALL",
                    maintenance_db="db_name",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )

                query = 'GRANT ALL ON TABLE public."awl" TO "baruwa"'

                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "db_name",
                        "-c",
                        query,
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

        # Test grant on all tables
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.has_privileges", Mock(return_value=False)
            ):
                ret = postgres.privileges_grant(
                    "baruwa",
                    "ALL",
                    "table",
                    "SELECT",
                    maintenance_db="db_name",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )

                query = 'GRANT SELECT ON ALL TABLES IN SCHEMA public TO "baruwa"'

                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "db_name",
                        "-c",
                        query,
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_privileges_grant_group(self):
        """
        Test granting privileges on group
        """
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.has_privileges", Mock(return_value=False)
            ):
                ret = postgres.privileges_grant(
                    "baruwa",
                    "admins",
                    "group",
                    grant_option=True,
                    maintenance_db="db_name",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )

                query = 'GRANT admins TO "baruwa" WITH ADMIN OPTION'

                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "db_name",
                        "-c",
                        query,
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.has_privileges", Mock(return_value=False)
            ):
                ret = postgres.privileges_grant(
                    "baruwa",
                    "admins",
                    "group",
                    maintenance_db="db_name",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )

                query = 'GRANT admins TO "baruwa"'

                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "db_name",
                        "-c",
                        query,
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_privileges_revoke_table(self):
        """
        Test revoking privileges on table
        """
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.has_privileges", Mock(return_value=True)):
                ret = postgres.privileges_revoke(
                    "baruwa",
                    "awl",
                    "table",
                    "ALL",
                    maintenance_db="db_name",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )

                query = "REVOKE ALL ON TABLE public.awl FROM baruwa"

                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "db_name",
                        "-c",
                        query,
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_privileges_revoke_group(self):
        """
        Test revoking privileges on group
        """
        with patch(
            "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
        ):
            with patch("salt.modules.postgres.has_privileges", Mock(return_value=True)):
                ret = postgres.privileges_revoke(
                    "baruwa",
                    "admins",
                    "group",
                    maintenance_db="db_name",
                    runas="user",
                    host="testhost",
                    port="testport",
                    user="testuser",
                    password="testpassword",
                )

                query = "REVOKE admins FROM baruwa"

                postgres._run_psql.assert_called_once_with(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "db_name",
                        "-c",
                        query,
                    ],
                    host="testhost",
                    port="testport",
                    password="testpassword",
                    user="testuser",
                    runas="user",
                )

    def test_datadir_init(self):
        """
        Test Initializing a postgres data directory
        """
        with patch(
            "salt.modules.postgres._run_initdb", Mock(return_value={"retcode": 0})
        ):
            with patch(
                "salt.modules.postgres.datadir_exists", Mock(return_value=False)
            ):
                name = "/var/lib/pgsql/data"
                ret = postgres.datadir_init(
                    name, user="postgres", password="test", runas="postgres"
                )
                postgres._run_initdb.assert_called_once_with(
                    name,
                    auth="password",
                    encoding="UTF8",
                    locale=None,
                    password="test",
                    runas="postgres",
                    checksums=False,
                    waldir=None,
                    user="postgres",
                )
                self.assertTrue(ret)

    def test_datadir_exists(self):
        """
        Test Checks if postgres data directory has been initialized
        """
        with patch("os.path.isfile", Mock(return_value=True)):
            name = "/var/lib/pgsql/data"
            ret = postgres.datadir_exists(name)
            self.assertTrue(ret)

    def test_pg_is_older_ext_ver(self):
        """
        Test Checks if postgres extension version string is older
        """
        self.assertTrue(postgres._pg_is_older_ext_ver("8.5", "9.5"))
        self.assertTrue(postgres._pg_is_older_ext_ver("8.5", "8.6"))
        self.assertTrue(postgres._pg_is_older_ext_ver("8.5.2", "8.5.3"))
        self.assertFalse(postgres._pg_is_older_ext_ver("9.5", "8.5"))
        self.assertTrue(postgres._pg_is_older_ext_ver("9.5", "9.6"))
        self.assertTrue(postgres._pg_is_older_ext_ver("9.5.0", "9.5.1"))
        self.assertTrue(postgres._pg_is_older_ext_ver("9.5", "9.5.1"))
        self.assertFalse(postgres._pg_is_older_ext_ver("9.5.1", "9.5"))
        self.assertFalse(postgres._pg_is_older_ext_ver("9.5b", "9.5a"))
        self.assertTrue(postgres._pg_is_older_ext_ver("10a", "10b"))
        self.assertTrue(postgres._pg_is_older_ext_ver("1.2.3.4", "1.2.3.5"))
        self.assertTrue(postgres._pg_is_older_ext_ver("10dev", "10next"))
        self.assertFalse(postgres._pg_is_older_ext_ver("10next", "10dev"))



# Late import so mock can do its job
import salt.states.gem as gem
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class TestGemState(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {gem: {"__opts__": {"test": False}}}

    def test_installed(self):
        gems = {"foo": ["1.0"], "bar": ["2.0"]}
        gem_list = MagicMock(return_value=gems)
        gem_install_succeeds = MagicMock(return_value=True)
        gem_install_fails = MagicMock(return_value=False)

        with patch.dict(gem.__salt__, {"gem.list": gem_list}):
            with patch.dict(gem.__salt__, {"gem.install": gem_install_succeeds}):
                ret = gem.installed("foo")
                self.assertEqual(True, ret["result"])
                ret = gem.installed("quux")
                self.assertEqual(True, ret["result"])
                gem_install_succeeds.assert_called_once_with(
                    "quux",
                    pre_releases=False,
                    ruby=None,
                    runas=None,
                    version=None,
                    proxy=None,
                    rdoc=False,
                    source=None,
                    ri=False,
                    gem_bin=None,
                )

            with patch.dict(gem.__salt__, {"gem.install": gem_install_fails}):
                ret = gem.installed("quux")
                self.assertEqual(False, ret["result"])
                gem_install_fails.assert_called_once_with(
                    "quux",
                    pre_releases=False,
                    ruby=None,
                    runas=None,
                    version=None,
                    proxy=None,
                    rdoc=False,
                    source=None,
                    ri=False,
                    gem_bin=None,
                )

    def test_installed_version(self):
        gems = {"foo": ["1.0"], "bar": ["2.0"]}
        gem_list = MagicMock(return_value=gems)
        gem_install_succeeds = MagicMock(return_value=True)

        with patch.dict(gem.__salt__, {"gem.list": gem_list}):
            with patch.dict(gem.__salt__, {"gem.install": gem_install_succeeds}):
                ret = gem.installed("foo", version=">= 1.0")
                self.assertEqual(True, ret["result"])
                self.assertEqual(
                    "Installed Gem meets version requirements.", ret["comment"]
                )

    def test_removed(self):
        gems = ["foo", "bar"]
        gem_list = MagicMock(return_value=gems)
        gem_uninstall_succeeds = MagicMock(return_value=True)
        gem_uninstall_fails = MagicMock(return_value=False)
        with patch.dict(gem.__salt__, {"gem.list": gem_list}):
            with patch.dict(gem.__salt__, {"gem.uninstall": gem_uninstall_succeeds}):
                ret = gem.removed("quux")
                self.assertEqual(True, ret["result"])
                ret = gem.removed("foo")
                self.assertEqual(True, ret["result"])
                gem_uninstall_succeeds.assert_called_once_with(
                    "foo", None, runas=None, gem_bin=None
                )

            with patch.dict(gem.__salt__, {"gem.uninstall": gem_uninstall_fails}):
                ret = gem.removed("bar")
                self.assertEqual(False, ret["result"])
                gem_uninstall_fails.assert_called_once_with(
                    "bar", None, runas=None, gem_bin=None
                )

    def test_sources_add(self):
        gem_sources = ["http://foo", "http://bar"]
        gem_sources_list = MagicMock(return_value=gem_sources)
        gem_sources_add_succeeds = MagicMock(return_value=True)
        gem_sources_add_fails = MagicMock(return_value=False)
        with patch.dict(gem.__salt__, {"gem.sources_list": gem_sources_list}):
            with patch.dict(
                gem.__salt__, {"gem.sources_add": gem_sources_add_succeeds}
            ):
                ret = gem.sources_add("http://foo")
                self.assertEqual(True, ret["result"])
                ret = gem.sources_add("http://fui")
                self.assertEqual(True, ret["result"])
                gem_sources_add_succeeds.assert_called_once_with(
                    source_uri="http://fui", ruby=None, runas=None
                )
            with patch.dict(gem.__salt__, {"gem.sources_add": gem_sources_add_fails}):
                ret = gem.sources_add("http://fui")
                self.assertEqual(False, ret["result"])
                gem_sources_add_fails.assert_called_once_with(
                    source_uri="http://fui", ruby=None, runas=None
                )

    def test_sources_remove(self):
        gem_sources = ["http://foo", "http://bar"]
        gem_sources_list = MagicMock(return_value=gem_sources)
        gem_sources_remove_succeeds = MagicMock(return_value=True)
        gem_sources_remove_fails = MagicMock(return_value=False)
        with patch.dict(gem.__salt__, {"gem.sources_list": gem_sources_list}):
            with patch.dict(
                gem.__salt__, {"gem.sources_remove": gem_sources_remove_succeeds}
            ):
                ret = gem.sources_remove("http://fui")
                self.assertEqual(True, ret["result"])
                ret = gem.sources_remove("http://foo")
                self.assertEqual(True, ret["result"])
                gem_sources_remove_succeeds.assert_called_once_with(
                    source_uri="http://foo", ruby=None, runas=None
                )
            with patch.dict(
                gem.__salt__, {"gem.sources_remove": gem_sources_remove_fails}
            ):
                ret = gem.sources_remove("http://bar")
                self.assertEqual(False, ret["result"])
                gem_sources_remove_fails.assert_called_once_with(
                    source_uri="http://bar", ruby=None, runas=None
                )



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>

    Test cases for salt.modules.oracle
"""


import os

import pytest

import salt.modules.oracle as oracle
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {oracle: {"cx_Oracle": object()}}


def test_run_query():
    """
    Test for Run SQL query and return result
    """
    with patch.object(oracle, "_connect", MagicMock()) as mock_connect:
        mock_connect.cursor.execute.fetchall.return_value = True
        with patch.object(oracle, "show_dbs", MagicMock()):
            assert oracle.run_query("db", "query")


def test_show_dbs():
    """
    Test for Show databases configuration from pillar. Filter by `*args`
    """
    with patch.dict(oracle.__salt__, {"pillar.get": MagicMock(return_value="a")}):
        assert oracle.show_dbs("A", "B") == {"A": "a", "B": "a"}

        assert oracle.show_dbs() == "a"


def test_version():
    """
    Test for Server Version (select banner  from v$version)
    """
    with patch.dict(oracle.__salt__, {"pillar.get": MagicMock(return_value="a")}):
        with patch.object(oracle, "run_query", return_value="A"):
            assert oracle.version() == {}


def test_client_version():
    """
    Test for Oracle Client Version
    """
    with patch.object(oracle, "cx_Oracle", MagicMock(side_effect=MagicMock())):
        assert oracle.client_version() == ""


def test_show_pillar():
    """
    Test for Show Pillar segment oracle.*
    """
    with patch.dict(oracle.__salt__, {"pillar.get": MagicMock(return_value="a")}):
        assert oracle.show_pillar("item") == "a"


def test_show_env():
    """
    Test for Show Environment used by Oracle Client
    """
    with patch.object(
        os,
        "environ",
        return_value={
            "PATH": "PATH",
            "ORACLE_HOME": "ORACLE_HOME",
            "TNS_ADMIN": "TNS_ADMIN",
            "NLS_LANG": "NLS_LANG",
        },
    ):
        assert oracle.show_env() == {}



"""
External pillar module for testing the contents of __opts__ as seen
by external pillar modules.

Returns a hash of the name of the pillar module as defined in
_virtual__ with the value __opts__
"""


import logging

# Set up logging
log = logging.getLogger(__name__)

# DRY up the name we use
MY_NAME = "ext_pillar_opts"


def __virtual__():
    log.debug("Loaded external pillar %s as %s", __name__, MY_NAME)
    return True


def ext_pillar(minion_id, pillar, *args):
    return {MY_NAME: __opts__}



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""
import logging

import pytest

import salt.modules.libcloud_compute as libcloud_compute
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

REQUIRED_LIBCLOUD_VERSION = "2.0.0"
try:
    import libcloud
    from libcloud.compute.base import (
        BaseDriver,
        KeyPair,
        Node,
        NodeImage,
        NodeLocation,
        NodeSize,
        NodeState,
        StorageVolume,
        StorageVolumeState,
        VolumeSnapshot,
    )

    if hasattr(libcloud, "__version__") and Version(libcloud.__version__) < Version(
        REQUIRED_LIBCLOUD_VERSION
    ):
        raise ImportError()
    logging.getLogger("libcloud").setLevel(logging.CRITICAL)
    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False


if HAS_LIBCLOUD:

    class MockComputeDriver(BaseDriver):
        def __init__(self):  # pylint: disable=W0231
            self._TEST_SIZE = NodeSize(
                id="test_id",
                name="test_size",
                ram=4096,
                disk=10240,
                bandwidth=100000,
                price=0,
                driver=self,
            )
            self._TEST_NODE = Node(
                id="test_id",
                name="test_node",
                state=NodeState.RUNNING,
                public_ips=["1.2.3.4"],
                private_ips=["2.3.4.5"],
                driver=self,
                size=self._TEST_SIZE,
                extra={"ex_key": "ex_value"},
            )
            self._TEST_LOCATION = NodeLocation(
                id="test_location", name="location1", country="Australia", driver=self
            )
            self._TEST_VOLUME = StorageVolume(
                id="vol1",
                name="vol_name",
                size=40960,
                driver=self,
                state=StorageVolumeState.AVAILABLE,
                extra={"ex_key": "ex_value"},
            )
            self._TEST_VOLUME_SNAPSHOT = VolumeSnapshot(
                id="snap1", size=80960, driver=self
            )
            self._TEST_IMAGE = NodeImage(
                id="image1",
                name="test_image",
                extra={"ex_key": "ex_value"},
                driver=self,
            )
            self._TEST_KEY_PAIR = KeyPair(
                name="test_key",
                fingerprint="abc123",
                public_key="pub123",
                private_key="priv123",
                driver=self,
                extra={"ex_key": "ex_value"},
            )

        def list_nodes(self):
            return [self._TEST_NODE]

        def list_sizes(self, location=None):
            if location:
                assert location.id == "test_location"
            return [self._TEST_SIZE]

        def list_locations(self):
            return [self._TEST_LOCATION]

        def reboot_node(self, node):
            assert node.id == "test_id"
            return True

        def destroy_node(self, node):
            assert node.id == "test_id"
            return True

        def list_volumes(self):
            return [self._TEST_VOLUME]

        def list_volume_snapshots(self, volume):
            assert volume.id == "vol1"
            return [self._TEST_VOLUME_SNAPSHOT]

        def create_volume(self, size, name, location=None, snapshot=None):
            assert size == 9000
            assert name == "test_new_volume"
            if location:
                assert location.country == "Australia"
            return self._TEST_VOLUME

        def create_volume_snapshot(self, volume, name=None):
            assert volume.id == "vol1"
            if name:
                assert name == "test_snapshot"
            return self._TEST_VOLUME_SNAPSHOT

        def attach_volume(self, node, volume, device=None):
            assert node.id == "test_id"
            assert volume.id == "vol1"
            if device:
                assert device == "/dev/sdc"
            return True

        def detach_volume(self, volume):
            assert volume.id == "vol1"
            return True

        def destroy_volume(self, volume):
            assert volume.id == "vol1"
            return True

        def destroy_volume_snapshot(self, snapshot):
            assert snapshot.id == "snap1"
            return True

        def list_images(self, location=None):
            if location:
                assert location.id == "test_location"
            return [self._TEST_IMAGE]

        def create_image(self, node, name, description=None):
            assert node.id == "test_id"
            return self._TEST_IMAGE

        def delete_image(self, node_image):
            return True

        def get_image(self, image_id):
            assert image_id == "image1"
            return self._TEST_IMAGE

        def copy_image(self, source_region, node_image, name, description=None):
            assert source_region == "us-east1"
            assert node_image.id == "image1"
            assert name == "copy_test"
            return self._TEST_IMAGE

        def list_key_pairs(self):
            return [self._TEST_KEY_PAIR]

        def get_key_pair(self, name):
            assert name == "test_key"
            return self._TEST_KEY_PAIR

        def create_key_pair(self, name):
            assert name == "test_key"
            return self._TEST_KEY_PAIR

        def import_key_pair_from_string(self, name, key_material):
            assert name == "test_key"
            assert key_material == "test_key_value"
            return self._TEST_KEY_PAIR

        def import_key_pair_from_file(self, name, key_file_path):
            assert name == "test_key"
            assert key_file_path == "/path/to/key"
            return self._TEST_KEY_PAIR

        def delete_key_pair(self, key_pair):
            assert key_pair.name == "test_key"
            return True

else:
    MockComputeDriver = object


@pytest.mark.skipif(not HAS_LIBCLOUD, reason="No libcloud installed")
@patch(
    "salt.modules.libcloud_compute._get_driver",
    MagicMock(return_value=MockComputeDriver()),
)
class LibcloudComputeModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        module_globals = {
            "__salt__": {
                "config.option": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                )
            }
        }
        if libcloud_compute.HAS_LIBCLOUD is False:
            module_globals["sys.modules"] = {"libcloud": MagicMock()}

        return {libcloud_compute: module_globals}

    def test_module_creation(self):
        client = libcloud_compute._get_driver("test")
        self.assertFalse(client is None)

    def _validate_node(self, node):
        self.assertEqual(node["name"], "test_node")
        self.assertEqual(node["id"], "test_id")
        self.assertEqual(node["private_ips"], ["2.3.4.5"])
        self.assertEqual(node["public_ips"], ["1.2.3.4"])
        self.assertEqual(node["size"]["name"], "test_size")

    def _validate_size(self, size):
        self.assertEqual(size["id"], "test_id")
        self.assertEqual(size["name"], "test_size")
        self.assertEqual(size["ram"], 4096)

    def _validate_location(self, location):
        self.assertEqual(location["id"], "test_location")
        self.assertEqual(location["name"], "location1")
        self.assertEqual(location["country"], "Australia")

    def _validate_volume(self, volume):
        self.assertEqual(volume["id"], "vol1")
        self.assertEqual(volume["name"], "vol_name")
        self.assertEqual(volume["size"], 40960)
        self.assertEqual(volume["state"], "available")
        self.assertEqual(volume["extra"], {"ex_key": "ex_value"})

    def _validate_volume_snapshot(self, volume):
        self.assertEqual(volume["id"], "snap1")
        self.assertEqual(volume["size"], 80960)

    def _validate_image(self, image):
        self.assertEqual(image["id"], "image1")
        self.assertEqual(image["name"], "test_image")
        self.assertEqual(image["extra"], {"ex_key": "ex_value"})

    def _validate_key_pair(self, key):
        self.assertEqual(key["name"], "test_key")
        self.assertEqual(key["fingerprint"], "abc123")
        self.assertEqual(key["extra"], {"ex_key": "ex_value"})

    def test_list_nodes(self):
        nodes = libcloud_compute.list_nodes("test")
        self.assertEqual(len(nodes), 1)
        self._validate_node(nodes[0])

    def test_list_sizes(self):
        sizes = libcloud_compute.list_sizes("test")
        self.assertEqual(len(sizes), 1)
        self._validate_size(sizes[0])

    def test_list_sizes_location(self):
        sizes = libcloud_compute.list_sizes("test", location_id="test_location")
        self.assertEqual(len(sizes), 1)
        self._validate_size(sizes[0])

    def test_list_locations(self):
        locations = libcloud_compute.list_locations("test")
        self.assertEqual(len(locations), 1)
        self._validate_location(locations[0])

    def test_reboot_node(self):
        result = libcloud_compute.reboot_node("test_id", "test")
        self.assertTrue(result)

    def test_reboot_node_invalid(self):
        with self.assertRaises(ValueError):
            libcloud_compute.reboot_node("foo_node", "test")

    def test_destroy_node(self):
        result = libcloud_compute.destroy_node("test_id", "test")
        self.assertTrue(result)

    def test_destroy_node_invalid(self):
        with self.assertRaises(ValueError):
            libcloud_compute.destroy_node("foo_node", "test")

    def test_list_volumes(self):
        volumes = libcloud_compute.list_volumes("test")
        self.assertEqual(len(volumes), 1)
        self._validate_volume(volumes[0])

    def test_list_volume_snapshots(self):
        volumes = libcloud_compute.list_volume_snapshots("vol1", "test")
        self.assertEqual(len(volumes), 1)
        self._validate_volume_snapshot(volumes[0])

    def test_create_volume(self):
        volume = libcloud_compute.create_volume(9000, "test_new_volume", "test")
        self._validate_volume(volume)

    def test_create_volume_in_location(self):
        volume = libcloud_compute.create_volume(
            9000, "test_new_volume", "test", location_id="test_location"
        )
        self._validate_volume(volume)

    def test_create_volume_snapshot(self):
        snapshot = libcloud_compute.create_volume_snapshot("vol1", "test")
        self._validate_volume_snapshot(snapshot)

    def test_create_volume_snapshot_named(self):
        snapshot = libcloud_compute.create_volume_snapshot(
            "vol1", "test", name="test_snapshot"
        )
        self._validate_volume_snapshot(snapshot)

    def test_attach_volume(self):
        result = libcloud_compute.attach_volume("test_id", "vol1", "test")
        self.assertTrue(result)

    def test_detatch_volume(self):
        result = libcloud_compute.detach_volume("vol1", "test")
        self.assertTrue(result)

    def test_destroy_volume(self):
        result = libcloud_compute.destroy_volume("vol1", "test")
        self.assertTrue(result)

    def test_destroy_volume_snapshot(self):
        result = libcloud_compute.destroy_volume_snapshot("vol1", "snap1", "test")
        self.assertTrue(result)

    def test_list_images(self):
        images = libcloud_compute.list_images("test")
        self.assertEqual(len(images), 1)
        self._validate_image(images[0])

    def test_list_images_in_location(self):
        images = libcloud_compute.list_images("test", location_id="test_location")
        self.assertEqual(len(images), 1)
        self._validate_image(images[0])

    def test_create_image(self):
        image = libcloud_compute.create_image("test_id", "new_image", "test")
        self._validate_image(image)

    def test_delete_image(self):
        result = libcloud_compute.delete_image("image1", "test")
        self.assertTrue(result)

    def test_get_image(self):
        image = libcloud_compute.get_image("image1", "test")
        self._validate_image(image)

    def test_copy_image(self):
        new_image = libcloud_compute.copy_image(
            "us-east1", "image1", "copy_test", "test"
        )
        self._validate_image(new_image)

    def test_list_key_pairs(self):
        keys = libcloud_compute.list_key_pairs("test")
        self.assertEqual(len(keys), 1)
        self._validate_key_pair(keys[0])

    def test_get_key_pair(self):
        key = libcloud_compute.get_key_pair("test_key", "test")
        self._validate_key_pair(key)

    def test_create_key_pair(self):
        key = libcloud_compute.create_key_pair("test_key", "test")
        self._validate_key_pair(key)

    def test_import_key_string(self):
        key = libcloud_compute.import_key_pair("test_key", "test_key_value", "test")
        self._validate_key_pair(key)

    def test_import_key_file(self):
        key = libcloud_compute.import_key_pair(
            "test_key", "/path/to/key", "test", key_type="FILE"
        )
        self._validate_key_pair(key)

    def test_delete_key_pair(self):
        result = libcloud_compute.delete_key_pair("test_key", "test")
        self.assertTrue(result)



"""
Proxy Minion to manage Cisco Nexus Switches (NX-OS) over the NX-API

.. versionadded:: 2019.2.0

Proxy module for managing Cisco Nexus switches via the NX-API.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net>
:maturity:   new
:platform:   any

Usage
=====

.. note::

    To be able to use this module you need to enable to NX-API on your switch,
    by executing ``feature nxapi`` in configuration mode.

    Configuration example:

    .. code-block:: bash

        switch# conf t
        switch(config)# feature nxapi

    To check that NX-API is properly enabled, execute ``show nxapi``.

    Output example:

    .. code-block:: bash

        switch# show nxapi
        nxapi enabled
        HTTPS Listen on port 443

.. note::

    NX-API requires modern NXOS distributions, typically at least 7.0 depending
    on the hardware. Due to reliability reasons it is recommended to run the
    most recent version.

    Check https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus7000/sw/programmability/guide/b_Cisco_Nexus_7000_Series_NX-OS_Programmability_Guide/b_Cisco_Nexus_7000_Series_NX-OS_Programmability_Guide_chapter_0101.html
    for more details.

Pillar
------

The ``nxos_api`` proxy configuration requires the following parameters in order
to connect to the network switch:

transport: ``https``
    Specifies the type of connection transport to use. Valid values for the
    connection are ``http``, and  ``https``.

host: ``localhost``
    The IP address or DNS host name of the connection device.

username: ``admin``
    The username to pass to the device to authenticate the NX-API connection.

password
    The password to pass to the device to authenticate the NX-API connection.

port
    The TCP port of the endpoint for the NX-API connection. If this keyword is
    not specified, the default value is automatically determined by the
    transport type (``80`` for ``http``, or ``443`` for ``https``).

timeout: ``60``
    Time in seconds to wait for the device to respond. Default: 60 seconds.

verify: ``True``
    Either a boolean, in which case it controls whether we verify the NX-API
    TLS certificate, or a string, in which case it must be a path to a CA bundle
    to use. Defaults to ``True``.

    When there is no certificate configuration on the device and this option is
    set as ``True`` (default), the commands will fail with the following error:
    ``SSLError: [SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed (_ssl.c:581)``.
    In this case, you either need to configure a proper certificate on the
    device (*recommended*), or bypass the checks setting this argument as ``False``
    with all the security risks considered.

    Check https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus3000/sw/programmability/6_x/b_Cisco_Nexus_3000_Series_NX-OS_Programmability_Guide/b_Cisco_Nexus_3000_Series_NX-OS_Programmability_Guide_chapter_01.html
    to see how to properly configure the certificate.

All the arguments may be optional, depending on your setup.

Proxy Pillar Example
--------------------

.. code-block:: yaml

    proxy:
      proxytype: nxos_api
      host: switch1.example.com
      username: example
      password: example
"""

import copy
import logging

# Import Salt modules
from salt.exceptions import SaltException

# -----------------------------------------------------------------------------
# proxy properties
# -----------------------------------------------------------------------------

__proxyenabled__ = ["nxos_api"]
# proxy name

# -----------------------------------------------------------------------------
# globals
# -----------------------------------------------------------------------------

__virtualname__ = "nxos_api"
log = logging.getLogger(__name__)
nxos_device = {}

# -----------------------------------------------------------------------------
# property functions
# -----------------------------------------------------------------------------


def __virtual__():
    """
    This Proxy Module is widely available as there are no external dependencies.
    """
    return __virtualname__


# -----------------------------------------------------------------------------
# proxy functions
# -----------------------------------------------------------------------------


def init(opts):
    """
    Open the connection to the Nexsu switch over the NX-API.

    As the communication is HTTP based, there is no connection to maintain,
    however, in order to test the connectivity and make sure we are able to
    bring up this Minion, we are executing a very simple command (``show clock``)
    which doesn't come with much overhead and it's sufficient to confirm we are
    indeed able to connect to the NX-API endpoint as configured.
    """
    proxy_dict = opts.get("proxy", {})
    conn_args = copy.deepcopy(proxy_dict)
    conn_args.pop("proxytype", None)
    opts["multiprocessing"] = conn_args.pop("multiprocessing", True)
    # This is not a SSH-based proxy, so it should be safe to enable
    # multiprocessing.
    try:
        rpc_reply = __utils__["nxos_api.rpc"]("show clock", **conn_args)
        # Execute a very simple command to confirm we are able to connect properly
        nxos_device["conn_args"] = conn_args
        nxos_device["initialized"] = True
        nxos_device["up"] = True
    except SaltException:
        log.error("Unable to connect to %s", conn_args["host"], exc_info=True)
        raise
    return True


def ping():
    """
    Connection open successfully?
    """
    return nxos_device.get("up", False)


def initialized():
    """
    Connection finished initializing?
    """
    return nxos_device.get("initialized", False)


def shutdown(opts):
    """
    Closes connection with the device.
    """
    log.debug("Shutting down the nxos_api Proxy Minion %s", opts["id"])


# -----------------------------------------------------------------------------
# callable functions
# -----------------------------------------------------------------------------


def get_conn_args():
    """
    Returns the connection arguments of the Proxy Minion.
    """
    conn_args = copy.deepcopy(nxos_device["conn_args"])
    return conn_args


def rpc(commands, method="cli", **kwargs):
    """
    Executes an RPC request over the NX-API.
    """
    conn_args = nxos_device["conn_args"]
    conn_args.update(kwargs)
    return __utils__["nxos_api.rpc"](commands, method=method, **conn_args)



import os

import pytest


@pytest.fixture(scope="module")
def inst_dir():
    return r"C:\custom_location"


@pytest.fixture(scope="module")
def install(inst_dir):
    pytest.helpers.clean_env(inst_dir)

    # Create an existing config
    pytest.helpers.existing_config()

    pytest.helpers.run_command([pytest.INST_BIN, "/S", f"/install-dir={inst_dir}"])
    yield
    pytest.helpers.clean_env(inst_dir)


def test_binaries_present(install, inst_dir):
    assert os.path.exists(rf"{inst_dir}\ssm.exe")


def test_config_present(install):
    assert os.path.exists(rf"{pytest.DATA_DIR}\conf\minion")


def test_config_correct(install):
    # The config file should be the existing config, unchanged
    expected = pytest.EXISTING_CONTENT

    with open(rf"{pytest.DATA_DIR}\conf\minion") as f:
        result = f.readlines()

    assert result == expected
